import sys
from polly import application_error_info as app_err_info
from polly import http_response_codes as http_res_codes

sys.tracebacklimit = 0


class RequestException(Exception):
    def __init__(self, title, detail=None):
        self.title = title
        self.detail = detail


class BaseExceptionError(Exception):
    """
    Base Exception class for v2 APIs.
    All custom exceptions are created by extending this class.
    Exception has 4 attributes corresponding to details sent in 'error' object
    in response JSON -
        status - http status code
        code - application specific error code
        title - title of error
        detail - details of error
    """

    def __init__(self, status, code, title, detail):
        Exception.__init__(self)
        self.title = title
        self.detail = detail

    def as_dict(self):
        return {"title": self.title, "detail": self.detail}

    def as_str(self):
        exception_str = "Exception Type : " + self.__class__.__name__
        exception_str += "\nTitle - " + self.title if self.title else ""
        exception_str += "\nDetails - " + self.detail if self.detail else ""
        return exception_str

    def __str__(self):
        return f"{self.__class__.__name__} ({self.title}): {self.detail}"


class ElasticException(Exception):
    def __str__(self):
        if self.detail:
            return f"{self.title}: {self.detail}"
        return self.title


class UnauthorizedException(Exception):
    def __str__(self):
        return "Expired or Invalid Token"


class UnfinishedQueryException(Exception):
    def __init__(self, query_id):
        self.query_id = query_id

    def __str__(self):
        return f'Query "{self.query_id}" has not finished executing'


class QueryFailedException(Exception):
    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return f"Query failed to execute\n\treason: {self.reason}"


class OperationFailedException(Exception):
    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return f"{self.reason}"


class InvalidPathException(Exception):
    def __str__(self):
        return "This path does not represent a file or a directory. Please try again."


class InvalidCohortPathException(Exception):
    def __str__(self):
        return "This path does not represent a Cohort. Please try again."


class InvalidCohortNameException(Exception):
    def __str__(self, cohort_name):
        return f"The identifier {cohort_name} does not represent a valid cohort name. Please try again."


class InvalidRepoException(Exception):
    def __init__(self, repo_name):
        self.repo_name = repo_name

    def __str__(self):
        return f"The repository : {self.repo_name} is not supported. Please contact Polly Support."


class InvalidDatasetException(Exception):
    def __str__(self):
        return "Dataset/s not added."


class InvalidCohortOperationException(Exception):
    def __str__(self):
        return "This operation is not valid as no cohort has been instantiated."


class EmptyCohortException(Exception):
    def __str__(self):
        return "There are no datasets in the cohort. Please try adding datasets using add_to_cohort() function."


class CohortEditException(Exception):
    def __str__(self):
        return "No parameter specified for editing in cohort"


class InvalidCohortMergeOperation(Exception):
    def __str__(self):
        return "Incorrect or blank parameter specified for merging in cohort"


class InvalidCohortAddition(Exception):
    def __str__(self):
        return "The repository type is not compatible with the cohort due to different file structure. Please try again."


class OutdatedCohortVersion(Exception):
    def __init__(self, version):
        self.version = version

    def __str__(self):
        return f"The Cohort version is outdated. Please try again with the new version VERSION-{self.version}."


class TechnicalFaultException(Exception):
    def __str__(self):
        return "Samples not downloaded due to a technical fault. Please check \
the arguments passed and try again. Contact Polly Support in case of repeated failure."


class InvalidDatatypeException(Exception):
    def __str__(self):
        return "Single Cell Datatype is not incorporated. Please contact Polly Support."


class MissingKeyException(Exception):
    def __init__(self, key):
        self.key = key

    def __str__(self):
        return f"Missing keys {self.key}"


class InvalidParameterException(Exception):
    def __init__(self, parameter):
        self.parameter = parameter

    def __str__(self):
        return f"Empty or Invalid Parameters = {self.parameter}."


class IncompatibleDataSource(Exception):
    def __init__(self, parameter):
        self.parameter = parameter

    def __str__(self):
        return f"The datasets cannot be added due to conflicting data sources = {self.parameter}."


class InvalidFormatException(Exception):
    def __str__(self):
        return "This file format is not supported."


class UnsupportedRepositoryException(Exception):
    def __str__(self):
        return "Report Generation feature is enabled for 'GEO' repository. Please try again."


class paramException(BaseExceptionError):
    detail = app_err_info.PARAM_EXCEPTION

    def __init__(self, title=None, detail=None):
        self.title = app_err_info.PARAM_EXCEPTION_TITLE
        if detail:
            self.detail = detail


class wrongParamException(BaseExceptionError):
    detail = app_err_info.WRONG_PARAMS_EXCEPTION

    def __init__(self, title=None, detail=None):
        self.title = app_err_info.WRONG_PARAMS_EXCEPTION_TITLE
        if detail:
            self.detail = detail


class apiErrorException(BaseExceptionError):
    detail = app_err_info.API_ERROR_EXCEPTION

    def __init__(self, title=None, detail=None):
        self.title = app_err_info.API_ERROR_EXCEPTION_TITLE
        if detail:
            self.detail = detail


class invalidDataException(BaseExceptionError):
    def __init__(self, title=None, detail=None):
        self.title = app_err_info.INCORRECT_DATA
        if detail:
            self.detail = detail


class invalidApiResponseException(BaseExceptionError):
    def __init__(self, title=None, detail=None):
        if title:
            self.title = title
        if detail:
            self.detail = detail


class BadRequestError(BaseExceptionError):
    status = http_res_codes.BAD_REQUEST
    code = app_err_info.BAD_REQUEST_CODE
    title = app_err_info.BAD_REQUEST_TITLE
    detail = app_err_info.BAD_REQUEST_DETAIL

    def __init__(self, code=None, status=None, title=None, detail=None):
        if code:
            self.code = code
        if status:
            self.status = status
        if title:
            self.title = title
        if detail:
            self.detail = detail


class AccessDeniedError(BaseExceptionError):
    status = http_res_codes.FORBIDDEN
    code = app_err_info.FORBIDDEN_CODE
    title = app_err_info.FORBIDDEN_TITLE
    detail = app_err_info.FORBIDDEN_DETAIL

    def __init__(self, title=None, detail=None):
        if title:
            self.title = title
        if detail:
            self.detail = detail


class ResourceNotFoundError(BaseExceptionError):
    status = http_res_codes.NOT_FOUND
    code = app_err_info.NOT_FOUND_CODE
    title = app_err_info.NOT_FOUND_TITLE
    detail = app_err_info.NOT_FOUND_DETAIL

    def __init__(self, title=None, detail=None):
        if title:
            self.title = title
        if detail:
            self.detail = detail


class InvalidSchemaJsonException(Exception):
    def __str__(self):
        return "Schema error in the json request file; update and try again"


class InvalidSyntaxForRequestException(Exception):
    def __str__(self):
        return "Invalid syntax for this request was provided"


class EmptyPayloadException(Exception):
    def __str__(self):
        return "The specified payload could not be found. Inspect and try again."


class InvalidJobFunctionParameterException(Exception):
    def __init__(self, parameter):
        self.parameter = parameter

    def __str__(self):
        return (
            f"The specified {self.parameter} could not be found. Inspect and try again."
        )


def error_handler(response):
    if has_error_message(response):
        title, detail = extract_json_api_error(response)
        if title == app_err_info.REPOSITORY_LOCKED:
            detail = app_err_info.REPOSITORY_LOCKED_DETAIL
        raise RequestException(title, detail)
    elif response.status_code == 401:
        raise UnauthorizedException

    response.raise_for_status()


def has_error_message(response):
    try:
        for key in response.json().keys():
            if key in {"error", "errors"}:
                return True
        return False
    except Exception:
        return False


def extract_json_api_error(response):
    error = response.json().get("error")
    if error is None:
        error = response.json().get("errors")[0]
    if "title" in error:
        title = error.get("title")
    if "detail" in error:
        detail = error.get("detail")
    return title, detail


def extract_error_message_details(error_response):
    error = error_response.json().get("error")
    if error is None:
        error = error_response.json().get("errors")[0]

    type_ = error.get("type", None) or error.get("code", None)
    reason = error.get("reason", None) or error.get("title", None)
    details = error.get("details", None) or error.get("detail", None)

    return type_, reason, details


def is_unfinished_query_error(exception):
    return isinstance(exception, UnfinishedQueryException)



from multiprocessing import AuthenticationError
import re
from polly import helpers
from polly.auth import Polly
from polly.errors import (
    InvalidParameterException,
    InvalidPathException,
    UnauthorizedException,
    InvalidJobFunctionParameterException,
    extract_json_api_error,
)
from polly.helpers import (
    get_user_details,
    parseInt,
)

# from polly.constants import MIN_REQUIRED_KEYS_FOR_JOBS, MACHINES_FOR_JOBS
from polly.constants import COMPUTE_ENV_VARIABLE
import polly.http_response_codes as http_codes
from polly import constants as const

import pandas as pd
import os
import json


class jobs:
    """
    The polly_jobs class contains functions which can be used to create, cancel and monitor polly jobs,

    ``Args:``
        |  ``token (str):`` token copy from polly.

    .. code::


            from polly.jobs import jobs
            # To use this class initialize a object like this if you are not authorised on polly
            jobs = jobs(token)

    If you are authorised then you can initialize object without token to know about :ref:`authentication <auth>`.
    """

    def __init__(
        self,
        token=None,
        env="",
        default_env="polly",
    ) -> None:
        # check if COMPUTE_ENV_VARIABLE present or not
        # if COMPUTE_ENV_VARIABLE, give priority
        env = helpers.get_platform_value_from_env(
            COMPUTE_ENV_VARIABLE, default_env, env
        )
        self.session = Polly.get_session(token, env=env)
        self.base_url = f"https://v2.api.{self.session.env}.elucidata.io"
        self.discover_url = f"https://api.discover.{self.session.env}.elucidata.io"

    # project_id -> workspace_id
    # job_file -> json file
    def submit_job(self, project_id: str, job_file: str) -> pd.DataFrame:
        """
        submits  a polly cli job in the given workspace

        Args:
            project_id (str) :  workspace id : str
            job_file (str) : required configuration json file path : str

        Returns:
            on success: dataframe with workspace id and job id
            failure: throw errors
        """
        parameter_check_dict = {}
        parameter_check_dict["project_id"] = project_id
        parameter_check_dict["job_file"] = job_file
        self._parameter_check_for_jobs(parameter_check_dict)
        if isinstance(project_id, int):
            project_id = str(project_id)

        try:
            with open(job_file, "r") as jobfile:
                jobData = json.load(jobfile)
                # self._validate_job_json(jobData)
        except Exception as err:
            raise err

        # TODO: add file validation for json
        jobData = self._add_secret_env_keys_to_jobData(jobData, project_id)
        # print("** added secret env variables *** Job data looks like this: ")
        # print(jobData)
        job_post_data = self._submit_job_to_polly(project_id, jobData)
        # print(job_post_data)
        submitted_job_id = job_post_data.json().get("data").get("job_id")
        project_id = job_post_data.json().get("data").get("project_id")
        submittedProject = {"Workspace ID": project_id, "Job ID": submitted_job_id}
        submittedProject_df = pd.DataFrame([submittedProject])
        return submittedProject_df

    def cancel_job(self, project_id: str, job_id: str):
        """
        cancel a polly job.

        Args:
            project_id (str) : workspace id : str
            job_id (str) : job id

        Raises:
            InvalidParameterException
            err:
        """
        parameter_check_dict = {}
        parameter_check_dict["project_id"] = project_id
        parameter_check_dict["job_id"] = job_id
        self._parameter_check_for_jobs(parameter_check_dict)
        if isinstance(project_id, int):
            project_id = str(project_id)

        self.base_url = f"https://v2.api.{self.session.env}.elucidata.io"
        self.jobUrl = f"{self.base_url}/projects/{project_id}/jobs/{job_id}"
        try:
            job_data = {}
            job_data = self._add_secret_env_keys_to_jobData(
                jobData=job_data, project_id=project_id
            )
            self._get_headers_for_job(job_data)
            # successfull -> returns 204
            postData = self.session.delete(self.jobUrl)
            if postData.status_code == 204:
                print("Cancelled job ID " + job_id + " successfully!")
            else:
                # handling this differently because postData.text is a str
                if postData.text:
                    postData_resp_json = json.loads(postData.text)
                else:
                    postData_resp_json = postData.json()
                error = postData_resp_json.get("error")
                if error is None:
                    error = postData_resp_json.get("errors", [])[0]
                if "detail" in error:
                    detail = error.get("detail", "")
                    print("Failed to cancel the job.: " + detail)
                else:
                    print("Failed to cancel the job.")
        except Exception as err:
            print("Failed to cancel the job.")
            raise err

    def job_status(self, project_id: str, job_id="", internalCalls=False) -> dict:
        """
        Get the status of a job given the rproject id and job id.
        if no job id give, gives status for all the jobs in the
        provided workspace

        Arguments:
            project_id (str) -- workspace id
            job_id (str) --  job id

        Keyword Arguments:
            internalCalls -- _description_ (default: {False})

        Returns:
            dataframe with job id, job name and status sorted as per created
            timestamp
        """
        parameter_check_dict = {}
        parameter_check_dict["project_id"] = project_id
        if job_id:
            parameter_check_dict["job_id"] = job_id
        self._parameter_check_for_jobs(parameter_check_dict)
        if isinstance(project_id, int):
            project_id = str(project_id)

        sortedJobsTemp = None
        sortedJobs = []
        self.jobUrl = f"{self.base_url}/projects/{project_id}/jobs/{job_id}"

        # TODO: Yogesh : from a job id if we can get the project id. -> improvement
        # TODO: job id can be a optional parameter -> MVP
        jobData = {}
        try:
            updated_jobData = self._add_secret_env_keys_to_jobData(jobData, project_id)
            self._get_headers_for_job(updated_jobData)
            postDatas = self._get_data_for_job_id(self.jobUrl)
            if postDatas.status_code != const.OK:
                self._handle_response_errors(postDatas)
            # list of data
            else:
                sortedJobsTemp = postDatas.json()
                sortedJobs.extend(sortedJobsTemp.get("data"))
                for i in range(len(sortedJobsTemp.get("data"))):
                    if sortedJobsTemp.get("links", "") and sortedJobsTemp.get(
                        "links", ""
                    ).get("next", ""):
                        jobUrl = f"{self.base_url}{sortedJobsTemp.json().get('links').get('next')}"
                        postDatas = self._get_data_for_job_id(self, jobUrl)
                        if postDatas.status_code != const.OK:
                            self._handle_response_errors(postDatas)
                        sortedJobsTemp = postDatas.json()
                        sortedJobs.extend(sortedJobsTemp.get("data"))
                    else:
                        continue
        except Exception as err:
            print("Not able to get the status of the Job(s)")
            raise err
        sortedJobs = sorted(sortedJobs, key=lambda d: d["attributes"]["created_ts"])
        if internalCalls:
            return sortedJobs
        status_result_df = self._generate_job_status_df(sortedJobs)
        # with pd.option_context(
        #     "display.max_rows", 800, "display.max_columns", 800, "display.width", 1200
        # ):
        #     print(status_result_df)
        return status_result_df

    def _is_valid_job_id(self, project_id: str, job_id: str):
        """
        checks if the job id provided is valid
        checks the jobid in the jobs in the workspace

        Arguments:
            project_id (int or str) -- workspace id
            job_id (str) -- job id

        Returns:
            valid (bool) -- True if job id is valid
        """
        valid = False
        try:
            jobs_list = self.job_status(project_id, internalCalls=True)
            job_id_list = []
            for jobs in jobs_list:
                job_id_list.append(jobs.get("attributes").get("job_id"))
            if job_id in job_id_list:
                valid = True
        except Exception as err:
            raise err
        return valid

    def _handle_response_errors(self, postDatas):
        """
        api to handle the response errors

        Arguments:
            postDatas (json) : api response json

        """
        if postDatas.status_code == http_codes.FORBIDDEN:
            raise AuthenticationError("User access is denied. Please contact admin")
        elif postDatas.status_code == http_codes.UNAUTHORIZED:
            raise UnauthorizedException("User is unauthorized to access this")
        elif postDatas.status_code == http_codes.NOT_FOUND:
            error_title, error_detail = extract_json_api_error(postDatas.text)
            raise Exception(error_detail)
        else:
            error_title, error_detail = extract_json_api_error(postDatas.text)
            raise Exception(error_detail)

    def _generate_job_status_df(self, sortedJobs):
        """
        given a list of jobs related info dict this generates a dataframe of the job id, name and status

        Arguments:
            sortedJobs (list): list of job information

        Returns:
            status_result_df (pd.dataframe): dataframe containing job id, name
        """
        job_info_list = []
        if sortedJobs:
            for jobs in sortedJobs:
                job_info = {}
                job_info["id"] = jobs.get("attributes").get("job_id")
                job_info["name"] = jobs.get("attributes").get("job_name")
                job_info["job_state"] = jobs.get("attributes").get("state")
                job_info_list.append(job_info)
            status_result_df = pd.DataFrame(job_info_list)
            status_result_df.columns = ["Job ID", "Job Name", "Job State"]
        else:
            status_result_df = pd.DataFrame(columns=["Job ID", "Job Name", "Job State"])
        return status_result_df

    def _get_data_for_job_id(self, jobUrl):
        """
        given a url, gets the response for the url.
        """
        try:
            self.jobUrl = jobUrl
            if jobUrl:
                postDatas = self.session.get(self.jobUrl)
            else:
                postDatas = self.session.get(self.jobUrl)
        except Exception as err:
            raise err
        return postDatas

    def _add_secret_env_keys_to_jobData(self, jobData: json, project_id: str) -> json:
        """
        add env variables and values to the job data json needed for job run

        Args:
            jobData (json) :  json containing existing variables
            project_id (str):  workspace id

        Returns:
           appended job data json with the secret env variables
        """
        dict_secret_env_variables = self._get_secret_env_variables()
        if "secret_env" not in jobData:
            jobData["secret_env"] = {}
        jobData["secret_env"]["POLLY_REFRESH_TOKEN"] = dict_secret_env_variables.get(
            "refreshToken"
        )
        jobData["secret_env"]["POLLY_ID_TOKEN"] = dict_secret_env_variables.get(
            "idToken"
        )
        jobData["secret_env"]["POLLY_WORKSPACE_ID"] = project_id
        jobData["secret_env"]["POLLY_USER"] = dict_secret_env_variables.get("email")
        jobData["secret_env"]["POLLY_SUB"] = dict_secret_env_variables.get("sub")
        jobData["secret_env"]["POLLY_EXP"] = dict_secret_env_variables.get("exp")
        jobData["secret_env"]["POLLY_AUD"] = dict_secret_env_variables.get("aud")

        return jobData

    def _submit_job_to_polly(self, project_id: str, job_data: json):
        """
        given a complete job data and workspace id
        calling the api to start a job

        Args:
            project_id (str) : workspace id
            job_data (jsob) : json containing all required variables

        Raises:
            err

        Returns:
            response from the api call
        """
        submitBody = {"data": {"type": "jobs", "attributes": job_data}}

        self.base_url = f"https://v2.api.{self.session.env}.elucidata.io"
        self.jobUrl = f"{self.base_url}/projects/{project_id}/jobs"

        self._get_headers_for_job(job_data)
        try:
            postData = self.session.post(self.jobUrl, data=json.dumps(submitBody))
            # print(postData)
            # print(postData.json())
            error = postData.json().get("error")
            if error is not None:
                error = postData.json().get("errors")[0]
                if "title" in error:
                    title = error.get("title")
                    print(title)
                if "detail" in error:
                    detail = error.get("detail")
                    print(detail)
        except Exception as err:
            print("Not able to submit job")
            raise err
        return postData

    def _get_headers_for_job(self, job_data: json):
        """
        adds required headers for polly job api request
        in the session headers itself

        Arguments:
            job_data (json): json
        """
        # print("**** in get headers for job.. adding cookie for job..")
        self.session.headers["Cookie"] = helpers.makeRequestCookieForPollyJob(job_data)

    """
    def _validate_job_json(self, jobData: json):
        # first check for if it is valid json format is done as part of json.load()
        # second check for // checking if necessary fiels are present
        min_required_feilds = MIN_REQUIRED_KEYS_FOR_JOBS
        # find the min_required_feilds missing from jobdata
        diff_req_keys = []
        for req_key in min_required_feilds:
            if req_key not in jobData.keys():
                diff_req_keys.append(req_key)
        if len(diff_req_keys) > 0:
            if not (
                ((not ("cpu" in diff_req_keys)) and not ("memory" in diff_req_keys))
                or not ("machineType" in diff_req_keys)
            ):
                raise InvalidParameterException(
                    "Please mention either cpu and memory or machine type in the .json file"
                )

            if "cpu" in diff_req_keys:
                diff_req_keys.remove("cpu")
            if "memory" in diff_req_keys:
                diff_req_keys.remove("memory")
            if "machineType" in diff_req_keys:
                diff_req_keys.remove("machineType")
            if len(diff_req_keys) > 0:
                raise InvalidParameterException(
                    "Following entry(s) are missing in the discription file: "
                    + ",".join(diff_req_keys)
                )
            if "machineType" in jobData.keys():
                resourceTypes = MACHINES_FOR_JOBS
                if jobData.get("machineType") not in resourceTypes.keys():
                    print("invalid machine type")
                    #TODO: handle this error and throw error message accordingly.
            if "cpu" in jobData.keys():
                if not self._valid_cpu(jobData["cpu"]):
                    raise InvalidParameterException(
                        "CPU given is not valid. Please check the documantation for more information."
                    )
                if (
                    self._convert_cpu(jobData.cpu) > 2
                    and "machineType" not in jobData.keys()
                ):
                    print(
                        "Job seems to need more resources. Please use machine type to specify resources. "
                    )
                    #TODO: handle this and throw error message accordingly.
        return
    """

    def _valid_cpu(self, cpu: str) -> bool:
        matches = False
        cpu_regex = "^([+-]?[0-9.]+)([m]*[-+]?[0-9]*)$"
        match = re.fullmatch(cpu_regex, cpu)
        if match is not None:
            matches = True
        return matches

    def _convert_cpu(self, cpu: str):
        cpuRegex = "^[0-9]+m$"
        if re.fullmatch(cpuRegex, cpu):
            cpu_parsed = parseInt(cpu[0:-1]) * 0.001
            return cpu_parsed
        else:
            return parseInt(cpu)

    def _get_secret_env_variables(self) -> dict:
        # print("*** in get secret env variables.. ***")
        dict_secret_env_variables = {}
        user_details = get_user_details(self)
        # print("********** CALLING helpers.get_user_details_using_aws_cognito(self).....")
        user_details_aws_cognito = helpers.get_user_details_using_aws_cognito(self)
        dict_secret_env_variables["id"] = user_details.get("id")
        dict_secret_env_variables["type"] = user_details.get("type")
        dict_secret_env_variables["user_id"] = user_details.get("user_id")
        dict_secret_env_variables["email"] = user_details.get("email")
        dict_secret_env_variables["first_name"] = user_details.get("first_name")
        dict_secret_env_variables["last_name"] = user_details.get("last_name")
        dict_secret_env_variables["active"] = user_details.get("last_name")
        dict_secret_env_variables["created_at"] = user_details.get("created_at")
        dict_secret_env_variables["organizations"] = user_details.get("organizations")
        dict_secret_env_variables["sub"] = user_details_aws_cognito.get("sub")
        dict_secret_env_variables["exp"] = user_details_aws_cognito.get("exp")
        dict_secret_env_variables["refreshToken"] = user_details_aws_cognito.get(
            "refresh_token"
        )
        dict_secret_env_variables["idToken"] = user_details_aws_cognito.get("id_token")
        dict_secret_env_variables["attributes"] = user_details
        dict_secret_env_variables["aud"] = user_details_aws_cognito.get("aud")

        return dict_secret_env_variables

    def _parameter_check_for_jobs(self, paramter_dict: dict):
        for keys, values in paramter_dict.items():
            if keys == "project_id":
                project_id = values
                if not (
                    project_id
                    and (isinstance(project_id, str) or isinstance(project_id, int))
                ):
                    raise InvalidParameterException("project id/workspace id")
            if keys == "job_file":
                job_file = values
                if not (job_file):
                    raise InvalidParameterException("job_file")
                if not os.path.exists(job_file):
                    raise InvalidPathException(job_file)
            if keys == "job_id":
                job_id = values
                if not (job_id and (isinstance(job_id, str))):
                    raise InvalidParameterException(
                        "Missing/invalid datatype for job id"
                    )

                if not self._is_valid_job_id(project_id, job_id):
                    raise InvalidJobFunctionParameterException("Job id")



from requests import Session
import os
import base64
import json


class PollySession(Session):
    """
    This class contain function to create session for
    polly.

    ``Args:``
        |  ``token (str):`` token copy from polly.

    ``Returns:``
        |  None

    To use this function


    .. code::


            from polly.session import PollySession
            session = PollySession(token)

    """

    def __init__(self, TOKEN, env="polly"):
        Session.__init__(self)
        try:
            # for python version >= python3.8
            from importlib.metadata import version

            version = version("polly-python")
        except ImportError:
            # for python version < python3.8
            import pkg_resources

            version = pkg_resources.get_distribution("polly-python").version
        client = os.getenv("POLLY_SERVICE")
        if client is not None:
            version = version + "/" + client
        else:
            version = version + "/local"

        self.token = TOKEN
        # check to differentiate between api auth key and refresh token
        # this check is based on the premise that

        # REFRESH TOKEN =>  It is not base64 encoded string.
        # There are 5 parts of REFRESH_TOKEN(all seperated by `.`)
        # Part 1 -> base64 encoded json string-> contains cty(technology of token),
        # algorithm used to encode and encryption type
        # Part 2,Part 3, Part 4, Part 5 -> Not base64 encoded json.

        # API AUTH KEY => base64 encoded string, no `.` in between.

        # check applied
        # split the token by `.` and take the 1st part
        # if the 1st part is a base64 encoded json string
        # that means its a refresh token
        # because API AUTH KEY is if we decode it is not
        # base64 encoded json neither it has `.` in it.
        # decoding API AUTH KEY => gives decoded string

        try:
            json.loads(base64.b64decode(TOKEN.split(".")[0].encode("ascii") + b"=="))
            self.headers = {
                "Content-Type": "application/vnd.api+json",
                "Cookie": f"refreshToken={TOKEN}",
                "User-Agent": "polly-python/" + version,
            }
        except Exception:
            self.headers = {
                "Content-Type": "application/vnd.api+json",
                "x-api-key": f"{TOKEN}",
                "User-Agent": "polly-python/" + version,
            }
        self.env = env



import os
import json
import base64
from polly import helpers
from pathlib import Path
import logging
import pandas as pd
import collections
from polly.errors import (
    InvalidCohortOperationException,
    InvalidParameterException,
    InvalidPathException,
    EmptyCohortException,
    InvalidCohortPathException,
    OutdatedCohortVersion,
)
from polly.constants import OBSOLETE_METADATA_FIELDS, COHORT_VERSION, dot
import shutil
from joblib import Parallel, delayed
from cmapPy.pandasGEXpress.parse import parse
from cmapPy.pandasGEXpress.write_gct import write
from cmapPy.pandasGEXpress.concat import assemble_data
from cmapPy.pandasGEXpress import GCToo


class CohortFileStandard:
    """
    The Class responsible for Encapsulating Core Cohort Functionalities.
    """

    def __init__(self) -> None:
        self._cohort_details = None

    def load_cohort(self, local_path: str):
        """
        Function to load an existing cohort into an object.
        Once loaded, the functions described in the documentation can be used for the object where the cohort is loaded.
        Args:
            | local_path(str): local path of the cohort
        Returns:
            | A confirmation message on instantiation of the cohort
        """
        if not os.path.exists(local_path):
            raise InvalidPathException(local_path)
        file_meta = helpers.make_path(local_path, "cohort.meta")
        if not os.path.exists(file_meta):
            raise InvalidCohortPathException
        file = open(file_meta, "r")
        byte = file.read()
        file.close()
        data = base64.b64decode((byte))
        str_data = data.decode("utf-8")
        json_data = json.loads(str_data)
        version = json_data.get("version")
        if version != COHORT_VERSION:
            raise OutdatedCohortVersion(COHORT_VERSION)
        logging.basicConfig(level=logging.INFO)
        logging.info("Cohort Loaded !")
        self._cohort_details = json_data

    def summarize_cohort(self, cohort_path):
        """
        Function to return metadata and summary of a cohort
        Returns:
            | A tuple with the first value as cohort metadata information (name, description and number of dataset(s)
              or sample(s) in the cohort) and the second value as dataframe containing the source, dataset_id or sample_id
              and data type available in the cohort.
        """
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        meta_details = self._get_metadetails(cohort_path)
        df_details = None
        if "entity_type" not in self._cohort_details:
            df_details = self._get_single_df()
            return meta_details, df_details
        if self._cohort_details.get("entity_type") == "dataset":
            df_details = self._get_single_df()
        else:
            df_details = self._get_multiple_df()
        return meta_details, df_details

    def _get_metadetails(self, cohort_path) -> dict:
        """
        Function to return metadata details of a cohort
        """
        meta_dict = {}
        folder_name = os.path.basename(cohort_path)
        cohort_name = folder_name.split(".")[0]
        meta_dict["cohort_name"] = cohort_name
        meta_details = ["description", "number_of_datasets"]
        for key, value in self._cohort_details.items():
            if key.lower() in meta_details:
                meta_dict[key] = value
        return meta_dict

    def _get_single_df(self) -> pd.DataFrame:
        """
        Function to return cohort summary in a dataframe
        """
        df_dict = {"source_omixatlas": [], "datatype": [], "dataset_id": []}
        dataset_list = list(self._cohort_details["entity_id"].keys())
        for entity in dataset_list:
            omixatlas = self._cohort_details.get("entity_id", {}).get(entity, {})[0]
            data_type = self._cohort_details.get("entity_id", {}).get(entity, {})[1]
            df_dict["dataset_id"].append(entity)
            df_dict["source_omixatlas"].append(omixatlas)
            df_dict["datatype"].append(data_type)
        dataframe = pd.DataFrame.from_dict(df_dict)
        return dataframe

    def _get_multiple_df(self) -> pd.DataFrame:
        """
        Function to return cohort summary in a dataframe
        """
        df_dict = {
            "source_omixatlas": [],
            "datatype": [],
            "dataset_id": [],
            "sample_id": [],
            "number_of_samples": [],
        }
        dataset_list = list(self._cohort_details["entity_id"].keys())
        for entity in dataset_list:
            omixatlas = self._cohort_details.get("entity_id", {}).get(entity, {})[0]
            data_type = self._cohort_details.get("entity_id", {}).get(entity, {})[1]
            sample_id = (
                self._cohort_details.get("source_omixatlas", {})
                .get(omixatlas, {})
                .get(entity)
            )
            df_dict["dataset_id"].append(entity)
            df_dict["source_omixatlas"].append(omixatlas)
            df_dict["datatype"].append(data_type)
            df_dict["sample_id"].append(sample_id)
            df_dict["number_of_samples"].append(len(sample_id))
        dataframe = pd.DataFrame.from_dict(df_dict)
        return dataframe

    def _read_gcts(self, dataset_ids: list, file_path: str) -> list:
        gct_files = [
            f"{file_path}/{self._cohort_details.get('entity_id',{}).get(dataset_id)[0]}_{dataset_id}.gct"
            for dataset_id in dataset_ids
        ]
        results_gct = Parallel(n_jobs=len(gct_files))(
            delayed(parse)(gct_file) for gct_file in gct_files
        )
        return results_gct

    def merge_single_sample_metadata(self, cohort_path: str):
        """
        Function to merge the sample level metadata from all the gct files in a cohort.
        Returns:
            | A pandas dataframe containing the merged metadata for analysis.
        """
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        sample_list = list(self._cohort_details["entity_id"].keys())
        if len(sample_list) == 0:
            raise EmptyCohortException
        results_gct = self._read_gcts(sample_list, cohort_path)
        for i in range(0, len(results_gct)):
            df = results_gct[i].col_metadata_df
            index_l = len(df.index)
            new_dataset = [sample_list[i]] * index_l
            results_gct[i].col_metadata_df.insert(
                loc=0, column="dataset_id", value=new_dataset
            )
        All_Metadata = pd.concat([i.col_metadata_df for i in results_gct])
        return All_Metadata

    def merge_single_data_matrix(self, cohort_path: str) -> pd.DataFrame:
        """
        Function to merge the data-matrix level metadata from all the gct files in a cohort.
        Returns:
            | A pandas dataframe containing the merged data for analysis.
        """
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        sample_list = list(self._cohort_details["entity_id"].keys())
        if len(sample_list) == 0:
            raise EmptyCohortException
        results_gct = self._read_gcts(sample_list, cohort_path)
        All_data_matrix = assemble_data(
            [i.data_df for i in results_gct], concat_direction="horiz"
        )
        return All_data_matrix

    def merge_dataset_metadata(self, cohort_path: str) -> pd.DataFrame:
        """
        Function to merge the dataset level metadata from all the jpco files in a cohort.
        Returns:
            | A pandas dataframe containing the merged data for analysis.
        """
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        entity_list = list(self._cohort_details["entity_id"].keys())
        if len(entity_list) == 0:
            raise EmptyCohortException
        jpco_files = [
            f"{cohort_path}/{self._cohort_details.get('entity_id',{}).get(entity_id)[0]}_{entity_id}.jpco"
            for entity_id in entity_list
        ]
        df_dict = []
        for files in jpco_files:
            with open(files, "r") as infile:
                temp_df = json.load(infile)
                df_dict.append(temp_df["_source"])
        df = pd.json_normalize(df_dict)
        first_column = df.pop("dataset_id")
        df.insert(loc=0, column="dataset_id", value=first_column)
        for col in OBSOLETE_METADATA_FIELDS:
            if col in df.columns:
                del df[col]
        return df

    def merge_feature_metadata(self, cohort_path) -> pd.DataFrame:
        """
        Function to merge the feature level metadata from all the gct files in a cohort.
        Returns:
            | A pandas dataframe containing the merged data for analysis.
        """
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        sample_list = list(self._cohort_details["entity_id"].keys())
        if len(sample_list) == 0:
            raise EmptyCohortException
        results_gct = self._read_gcts(sample_list, cohort_path)
        for i in range(0, len(results_gct)):
            df = results_gct[i].row_metadata_df
            index_l = len(df.index)
            new_dataset = [sample_list[i]] * index_l
            results_gct[i].row_metadata_df.insert(
                loc=0, column="dataset_id", value=new_dataset
            )
        All_Metadata = pd.concat([i.row_metadata_df for i in results_gct])
        return All_Metadata

    def merge_multiple_sample_metadata(self, cohort_path):
        """
        Function to merge the sample level metadata from all the gct files in a cohort.
        Returns:
            | A pandas dataframe containing the merged metadata for analysis.
        """
        sample_list = list(self._cohort_details["entity_id"].keys())
        results_gct = self._read_gcts(sample_list, cohort_path)
        oa_list = []
        for i in sample_list:
            omixatlas = self._cohort_details.get("entity_id", {}).get(i, {})[0]
            oa_list.append(omixatlas)
        new_df_list = []
        for i in range(0, len(results_gct)):
            omixatlas_details = self._cohort_details.get("source_omixatlas")
            sample_details = omixatlas_details[oa_list[i]][sample_list[i]]
            df = results_gct[i].col_metadata_df
            new_df = df.loc[sample_details]
            new_df_list.append(new_df)
        All_Metadata = pd.concat([i for i in new_df_list])
        return All_Metadata

    def merge_multiple_data_matrix(self, cohort_path: str) -> pd.DataFrame:
        """
        Function to merge the data-matrix level metadata from all the gct files in a cohort.
        Returns:
            | A pandas dataframe containing the merged data for analysis.
        """
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        sample_list = list(self._cohort_details["entity_id"].keys())
        if len(sample_list) == 0:
            raise EmptyCohortException
        results_gct = self._read_gcts(sample_list, cohort_path)
        oa_list = []
        for i in sample_list:
            omixatlas = self._cohort_details.get("entity_id", {}).get(i, {})[0]
            oa_list.append(omixatlas)
        new_df_list = []
        for i in range(0, len(results_gct)):
            omixatlas_details = self._cohort_details.get("source_omixatlas")
            sample_details = omixatlas_details[oa_list[i]][sample_list[i]]
            df = results_gct[i].data_df
            ripped_df = df[sample_details]
            new_df_list.append(ripped_df)
        All_data_matrix = pd.concat([i for i in new_df_list])
        return All_data_matrix

    def edit_cohort(self, new_cohort_name, new_description, cohort_path):
        returned_path_name = None
        # calling functions to edit description and then cohort name
        if new_description:
            self._edit_cohort_description(cohort_path, new_description)
        if new_cohort_name:
            returned_path_name = self._edit_cohort_name(cohort_path, new_cohort_name)
        return returned_path_name

    def _edit_cohort_name(self, cohort_path, new_cohort_name: str):
        """
        Function to edit cohort name
        """
        if not (new_cohort_name and isinstance(new_cohort_name, str)):
            return
        if dot in new_cohort_name:
            logging.error("The cohort name is not valid. Please try again.")
            return
        p = Path(cohort_path)
        parent = p.parent
        str_parent = str(parent.resolve())
        new_path = helpers.make_path(str_parent, f"{new_cohort_name}.pco")
        existing_path = cohort_path
        os.rename(existing_path, new_path)
        logging.basicConfig(level=logging.INFO)
        logging.info("Cohort Name Updated!")
        return new_path

    def _edit_cohort_description(self, cohort_path: str, new_description: str):
        """
        Function to edit cohort description
        """
        if not (new_description and isinstance(new_description, str)):
            return
        meta_path = helpers.make_path(cohort_path, "cohort.meta")
        json_data = None
        with open(meta_path, "r+b") as openfile:
            byte = openfile.read()
            data = base64.b64decode((byte))
            json_data = json.loads(data.decode("utf-8"))
            json_data["description"] = new_description
            input = json.dumps(json_data)
            encoded_data = base64.b64encode(input.encode("utf-8"))
            openfile.seek(0)
            openfile.write(encoded_data)
            openfile.truncate()
        logging.basicConfig(level=logging.INFO)
        logging.info("Cohort Description Updated!")
        self._cohort_details = json_data

    def is_valid(self, cohort_path: str) -> bool:
        """
        This function is used to check if a cohort is valid or not.
        Returns:
            | A boolean result based on the validity of the cohort.
        """
        if not os.path.exists(cohort_path):
            raise InvalidPathException
        meta_path = helpers.make_path(cohort_path, "cohort.meta")
        if not os.path.exists(meta_path):
            return False
        sample_list = list(self._cohort_details["entity_id"].keys())
        if len(sample_list) == 0:
            return True
        for sample in sample_list:
            omixatlas = self._cohort_details.get("entity_id", {}).get(sample)[0]
            gct_path = f"{cohort_path}/{omixatlas}_{sample}.gct"
            jpco_path = f"{cohort_path}/{omixatlas}_{sample}.jpco"
            if not (os.path.exists(gct_path) and os.path.exists(jpco_path)):
                return False
        return True

    def remove_single_from_cohort(self, cohort_path: str, dataset_id: list) -> None:
        """
        This function is used for removing dataset_id(s) from a cohort
        Args:
            | entity_id(list): list of dataset_id or sample_id to be removed from the cohort.
        Returns:
            | A confirmation message on removal of dataset_id or sample_id from cohort.
        """
        if not (dataset_id and isinstance(dataset_id, list)):
            raise InvalidParameterException("dataset_id")
        dataset_count = 0
        verified_dataset = []
        file_meta = helpers.make_path(cohort_path, "cohort.meta")
        dataset_id = list(set(dataset_id))
        json_data = None
        with open(file_meta, "r+b") as openfile:
            byte = openfile.read()
            data = base64.b64decode((byte))
            json_data = json.loads(data.decode("utf-8"))
            dataset_id_metadata = list(json_data["entity_id"].keys())
            for dataset in dataset_id:
                if dataset not in dataset_id_metadata:
                    logging.basicConfig(level=logging.INFO)
                    logging.info(f"Dataset Id - {dataset} not present in the Cohort.")
                    continue
                dataset_count += 1
                verified_dataset.append(dataset)
                omixatlas = json_data.get("entity_id", {}).get(dataset)[0]
                gct_path = f"{cohort_path}/{omixatlas}_{dataset}.gct"
                json_path = f"{cohort_path}/{omixatlas}_{dataset}.jpco"
                os.remove(gct_path)
                os.remove(json_path)
                del json_data.get("entity_id")[dataset]
                json_data.get("source_omixatlas").get(omixatlas).remove(dataset)
            omixatlas_dict = json_data.get("source_omixatlas")
            empty_keys = []
            for key, value in omixatlas_dict.items():
                if value == []:
                    empty_keys.append(key)
            for key in empty_keys:
                del omixatlas_dict[key]
            json_data["number_of_datasets"] -= dataset_count
            json_data["source_omixatlas"] = omixatlas_dict
            if not bool(json_data.get("entity_id")):
                if "entity_type" in json_data:
                    del json_data["entity_type"]
            input = json.dumps(json_data)
            encoded_data = base64.b64encode(input.encode("utf-8"))
            openfile.seek(0)
            openfile.write(encoded_data)
            openfile.truncate()
        logging.basicConfig(level=logging.INFO)
        logging.info(f"'{dataset_count}' dataset/s removed from Cohort!")
        self._cohort_details = json_data

    def remove_multiple_from_cohort(self, cohort_path, dataset_id: str, sample_id=[]):
        """
        This function is used for removing dataset_id or sample_id from a cohort
        Args:
            | entity_id(list): list of dataset_id or sample_id to be removed from the cohort.
        Returns:
            | A confirmation message on removal of dataset_id or sample_id from cohort.
        """
        if not (dataset_id and isinstance(dataset_id, str)):
            raise InvalidParameterException("dataset_id")
        file_meta = helpers.make_path(cohort_path, "cohort.meta")
        json_data = None
        with open(file_meta, "r+b") as openfile:
            byte = openfile.read()
            data = base64.b64decode((byte))
            json_data = json.loads(data.decode("utf-8"))
            dataset_id_metadata = list(json_data["entity_id"].keys())
            if dataset_id not in dataset_id_metadata:
                logging.basicConfig(level=logging.INFO)
                logging.info(f"Dataset Id - {dataset_id} not present in the Cohort.")
            else:
                source_omixatlas = json_data.get("source_omixatlas")
                omixatlas = json_data.get("entity_id", {}).get(dataset_id)[0]
                list_of_samples = source_omixatlas.get(omixatlas, {}).get(dataset_id)
                if len(sample_id) > 0:
                    sample_id = list(set(sample_id))
                    new_list_of_samples = [
                        x for x in list_of_samples if x not in sample_id
                    ]
                    if len(new_list_of_samples) != 0:
                        if collections.Counter(
                            new_list_of_samples
                        ) == collections.Counter(list_of_samples):
                            print("Invalid Samples to remove! ")
                        else:
                            json_data.get("source_omixatlas").get(omixatlas, {})[
                                dataset_id
                            ] = new_list_of_samples
                            print("Samples removed successfully!")
                    else:
                        json_data = self._remove_dataset(
                            dataset_id, json_data, cohort_path
                        )
                else:
                    json_data = self._remove_dataset(dataset_id, json_data, cohort_path)
                input = json.dumps(json_data)
                encoded_data = base64.b64encode(input.encode("utf-8"))
                openfile.seek(0)
                openfile.write(encoded_data)
                openfile.truncate()
        self._cohort_details = json_data

    def _remove_dataset(self, dataset_id: str, json_data: dict, cohort_path: str):
        omixatlas = json_data.get("entity_id", {}).get(dataset_id)[0]
        gct_path = f"{cohort_path}/{omixatlas}_{dataset_id}.gct"
        json_path = f"{cohort_path}/{omixatlas}_{dataset_id}.jpco"
        os.remove(gct_path)
        os.remove(json_path)
        del json_data.get("entity_id")[dataset_id]
        json_data.get("source_omixatlas").get(omixatlas).pop(dataset_id)
        omixatlas_dict = json_data.get("source_omixatlas")
        empty_keys = []
        for key, value in omixatlas_dict.items():
            if value == {}:
                empty_keys.append(key)
        for key in empty_keys:
            del omixatlas_dict[key]
        json_data["number_of_datasets"] -= 1
        json_data["source_omixatlas"] = omixatlas_dict
        if not bool(json_data.get("entity_id")):
            if "entity_type" in json_data:
                del json_data["entity_type"]
        logging.basicConfig(level=logging.INFO)
        logging.info(" 1 dataset removed from Cohort!")
        return json_data

    def _update_metadata_single(
        self, repo_name, entity_type, dataset_id, deleted_id, file_meta, dataset_tuple
    ):
        json_data = None
        with open(file_meta, "r+b") as openfile:
            byte = openfile.read()
            data = base64.b64decode((byte))
            json_data = json.loads(data.decode("utf-8"))
            source_omixatlas = json_data.get("source_omixatlas")
            if "entity_type" not in json_data:
                json_data["entity_type"] = entity_type
            if repo_name not in source_omixatlas:
                source_omixatlas[repo_name] = dataset_id
            else:
                [source_omixatlas[repo_name].append(i) for i in dataset_id]
            for i in range(0, len(dataset_id)):
                json_data["entity_id"][dataset_id[i]] = dataset_tuple[i]
                json_data["number_of_datasets"] += 1
            input = json.dumps(json_data)
            encoded_data = base64.b64encode(input.encode("utf-8"))
            openfile.seek(0)
            openfile.write(encoded_data)
            openfile.truncate()
        logging.basicConfig(level=logging.INFO)
        if deleted_id:
            logging.info("The following entities were not added : ")
            for id in deleted_id:
                print(f"{id}\n")
        logging.info(f"'{len(dataset_id)}' dataset/s added to Cohort!")
        return json_data

    def _update_metadata_multiple(
        self,
        repo_name,
        entity_type,
        dataset_id,
        sample_id,
        download_flag,
        file_meta,
        status_sample_id,
        dataset_tuple,
    ):
        json_data = None
        with open(file_meta, "r+b") as openfile:
            byte = openfile.read()
            data = base64.b64decode((byte))
            json_data = json.loads(data.decode("utf-8"))
            source_omixatlas = json_data.get("source_omixatlas")
            list_of_samples = []
            if "entity_type" not in json_data:
                json_data["entity_type"] = entity_type
            if repo_name not in source_omixatlas:
                source_omixatlas[repo_name] = {}
            elif dataset_id in source_omixatlas.get(repo_name):
                list_of_samples = source_omixatlas.get(repo_name, {}).get(dataset_id)
            sample_added = 0
            for i in range(0, len(sample_id)):
                if status_sample_id[i] and sample_id[i] not in list_of_samples:
                    sample_added += 1
                    list_of_samples.append(sample_id[i])
            source_omixatlas.get(repo_name, {})[dataset_id] = list_of_samples
            json_data["entity_id"][dataset_id] = dataset_tuple
            if download_flag:
                json_data["number_of_datasets"] += 1
            json_data["source_omixatlas"] = source_omixatlas
            input = json.dumps(json_data)
            encoded_data = base64.b64encode(input.encode("utf-8"))
            openfile.seek(0)
            openfile.write(encoded_data)
            openfile.truncate()
        logging.basicConfig(level=logging.INFO)
        logging.info(f"'{sample_added}' sample/s added to Cohort!")
        return json_data

    def _rollback_files(
        self,
        status_gct: list,
        status_jpco: list,
        repo_name: str,
        local_path: str,
        dataset_id: list,
    ) -> tuple:
        """
        Returns a list of dataset_ids and deleted_ids
        """
        deleted_id = []
        for i in status_gct:
            if i != 0:
                gct_file = f"{repo_name}_{i}.gct"
                gct_path = helpers.make_path(local_path, gct_file)
                jpco_file = f"{repo_name}_{i}.jpco"
                jpco_path = helpers.make_path(local_path, jpco_file)
                if os.path.exists(gct_path):
                    os.remove(gct_path)
                if os.path.exists(jpco_path):
                    os.remove(jpco_path)
                deleted_id.append(i)
                dataset_id.remove(i)
        for i in status_jpco:
            if i != 0 and i not in deleted_id:
                gct_file = f"{repo_name}_{i}.gct"
                gct_path = helpers.make_path(local_path, gct_file)
                jpco_file = f"{repo_name}_{i}.jpco"
                jpco_path = helpers.make_path(local_path, jpco_file)
                if os.path.exists(gct_path):
                    os.remove(gct_path)
                if os.path.exists(jpco_path):
                    os.remove(jpco_path)
                deleted_id.append(i)
                dataset_id.remove(i)
        return dataset_id, deleted_id

    def delete_cohort(self, cohort_path: str) -> None:
        """
        This function is used to delete a cohort.
        Returns:
            | A confirmation message on deletion of cohort
        """
        shutil.rmtree(cohort_path, ignore_errors=True)
        logging.basicConfig(level=logging.INFO)
        logging.info("Cohort Deleted Successfuly!")
        self._cohort_details = None

    def _rename_samples(self, dataset_id: str, gct_object: GCToo) -> GCToo:
        """
        Adds dataset_id as prefix to the sample names (cids) in the gct object
        """
        gct_object.data_df.columns = [
            dataset_id + "_" + sample for sample in gct_object.data_df.columns
        ]
        gct_object.col_metadata_df.index = [
            [dataset_id + "_" + sample for sample in gct_object.col_metadata_df.index]
        ]
        return gct_object

    def merge_gcts(
        self, entity_type: str, cohort_path: str, file_name: str, file_path: str
    ) -> None:
        """
        Merge the input gcts together to create one gct object. Only the features(genes) common to all the gcts are retained.
        """
        dataset_ids = list(self._cohort_details["entity_id"].keys())
        omixatlas_details = self._cohort_details.get("source_omixatlas")
        gct_files = [
            f"{cohort_path}/{self._cohort_details.get('entity_id',{}).get(dataset_id)[0]}_{dataset_id}.gct"
            for dataset_id in dataset_ids
        ]
        gct_objects = {
            dataset_id: parse(gct_file)
            for dataset_id, gct_file in zip(dataset_ids, gct_files)
        }
        # Creating a dict for dataset_id to samples mapping with dataset_id prefixed
        dataset_sample_mapping = {}
        # Collecting samples by differentiating between entity_types of cohort
        if entity_type == "sample":
            # Cohort contains repositories with multiple mapped datasets
            for dataset in dataset_ids:
                omixatlas = self._cohort_details.get("entity_id", {}).get(dataset, {})[
                    0
                ]
                samples = omixatlas_details.get(omixatlas, {}).get(dataset)
                dataset_sample_mapping[dataset] = [
                    f"{dataset}_{sample}" for sample in samples
                ]
        else:
            # Cohort contains repositories with single mapped datasets
            for dataset in dataset_ids:
                samples = list(gct_objects[dataset].col_metadata_df.index)
                dataset_sample_mapping[dataset] = [
                    f"{dataset}_{sample}" for sample in samples
                ]
        # Dictionaries for parsed data
        data_df_dict = {}
        col_metadata_dict = {}
        for dataset_id in dataset_ids:
            # Rename all samples for each gct and add dataset_id as a column to col metadata of each gct object
            gct_objects[dataset_id] = self._rename_samples(
                dataset_id, gct_objects[dataset_id]
            )
            gct_objects[dataset_id].col_metadata_df["dataset_id"] = dataset_id
            # Separating data_df and col_metadata_df, with tailored index and columns according to samples
            samples = dataset_sample_mapping[dataset_id]
            sampled_data_df = gct_objects[dataset_id].data_df[samples]
            sampled_col_metadata_df = gct_objects[dataset_id].col_metadata_df.loc[
                samples
            ]
            # Step to ensure index is a proper list
            sampled_col_metadata_df.index = samples
            data_df_dict[dataset_id] = sampled_data_df
            col_metadata_dict[dataset_id] = sampled_col_metadata_df

        if len(dataset_ids) == 1:
            # No need to merge as only one dataset present
            merged_data_df = data_df_dict[dataset_ids[0]]
            merged_col_metadata_df = col_metadata_dict[dataset_ids[0]]
            merged_row_metadata_df = gct_objects[dataset_ids[0]].row_metadata_df.loc[
                merged_data_df.index
            ]
        else:
            merged_data_df = pd.merge(
                data_df_dict[dataset_ids[0]],
                data_df_dict[dataset_ids[1]],
                left_index=True,
                right_index=True,
            )
            merged_col_metadata_df = pd.concat(
                [
                    col_metadata_dict[dataset_ids[0]],
                    col_metadata_dict[dataset_ids[1]],
                ],
                axis=0,
            )
            if len(dataset_ids) > 2:
                for dataset_id in dataset_ids[2:]:
                    merged_data_df = pd.merge(
                        merged_data_df,
                        data_df_dict[dataset_id],
                        left_index=True,
                        right_index=True,
                    )
                    merged_col_metadata_df = pd.concat(
                        [merged_col_metadata_df, col_metadata_dict[dataset_id]],
                        axis=0,
                    )
            merged_row_metadata_df = gct_objects[dataset_ids[0]].row_metadata_df.loc[
                merged_data_df.index
            ]
        merged_gct_object = GCToo.GCToo(
            data_df=merged_data_df,
            row_metadata_df=merged_row_metadata_df,
            col_metadata_df=merged_col_metadata_df,
        )
        write(merged_gct_object, helpers.make_path(file_path, file_name))



from polly.auth import Polly
from polly.errors import (
    InvalidParameterException,
    error_handler,
    InvalidPathException,
    AccessDeniedError,
)
from polly import helpers
from polly import constants as const
import logging
import pandas as pd
import json
import os
from polly.help import example, doc


class Workspaces:
    """
    This class contains functions to interact with workspaces on Polly. Users can create a workspace, fetch list\
 of workspaces, upload data to workspace and download data from workspace. To get started, users need to \
initialize a object that can use all function and methods of Workspaces class.

    ``Args:``
        |  ``token (str):`` token copy from polly.


    .. code::


            # to initialize a obj
            workspaces = Workspaces(token)
            # from there you can other methods
            workspaces.fetch_my_workspaces()

    If you are authorised then you can initialize object without token to know about \
:ref:`authentication <auth>`.
    """

    example = classmethod(example)
    doc = classmethod(doc)

    def __init__(self, token=None, env="", default_env="polly") -> None:
        # check if COMPUTE_ENV_VARIABLE present or not
        # if COMPUTE_ENV_VARIABLE, give priority
        env = helpers.get_platform_value_from_env(
            const.COMPUTE_ENV_VARIABLE, default_env, env
        )
        self.session = Polly.get_session(token, env=env)
        self.base_url = f"https://v2.api.{self.session.env}.elucidata.io"
        self.resource_url = f"{self.base_url}/workspaces"
        if self.session.env == "polly":
            self.env_string = "prod"
        elif self.session.env == "testpolly":
            self.env_string = "test"
        else:
            self.env_string = "devenv"

    def create_workspace(self, name: str, description=None):
        """
        This function create workspace on Polly.

        ``Args:``
            |  ``name (str):`` name of the workspace.
            |  ``description (str):`` general information about workspace.

        ``Returns:``
            |  It will return a object like this.
            .. code::



                        {
                        'id': 9999,
                        'name': 'rrrrr',
                        'active': True,
                        'description': 'for docu',
                        'created_time': '2022-03-16 11:08:47.127260',
                        'last_modified': '2022-03-16 11:08:47.127260',
                        'creator': 1127,
                        'project_property': {
                            'type': 'workspaces',
                            'labels': ''
                        },
                        'organisation': 1
                        }


        .. code::


                # to create a obj
                workspaces = Workspaces(token)
                # from there you can other methods
                workspaces.create_workspace("name_of_workspace")

        """
        url = self.resource_url
        payload = {
            "data": {
                "type": "workspaces",
                "attributes": {
                    "name": name,
                    "description": description,
                    "project_property": {"type": "workspaces", "labels": ""},
                },
            }
        }
        response = self.session.post(url, data=json.dumps(payload))
        error_handler(response)
        attributes = response.json()["data"]["attributes"]
        logging.basicConfig(level=logging.INFO)
        logging.info("Workspace Created !")
        return attributes

    def fetch_my_workspaces(self):
        """
        This function fetch workspaces from Polly.

        ``Args:``
            |  None

        ``Returns:``
            |  it will return a table with attributes.


        .. code::


                # create a obj
                workspaces = Workspaces(token)
                # from there you can other methods
                workspaces.fetch_my_workspaces()
        """
        all_details = self._fetch_workspaces_iteratively()
        pd.set_option("display.max_columns", 20)
        dataframe = pd.DataFrame.from_dict(
            pd.json_normalize(all_details), orient="columns"
        )
        dataframe.rename(
            columns={"id": "Workspace_id", "name": "Workspace_name"}, inplace=True
        )
        df = dataframe.sort_values(by="last_modified", ascending=False)
        df = df.reset_index()
        df.drop("index", axis=1, inplace=True)
        return df

    def _fetch_workspaces_iteratively(self):
        """
        Fetch all workspaces iteratively by making api calls until links is None.
        """
        url = self.resource_url
        all_details = []
        while True:
            response = self.session.get(url)
            error_handler(response)
            data = response.json()["data"]
            for workspace_details in data:
                self._modify_data(workspace_details)
                all_details.append(workspace_details.get("attributes"))
            links = response.json().get("links").get("next")
            if links is None:
                break
            url = f"{self.base_url}{links}"
        return all_details

    def _modify_data(self, data):
        """
        Removing less informative fields and making information user friendly
        """
        data_dict = data.get("attributes")
        if "project_property" in data_dict:
            del data_dict["project_property"]
        if "active" in data_dict:
            del data_dict["active"]
        if "created_time" in data_dict:
            del data_dict["created_time"]
        if "creator" in data_dict:
            del data_dict["creator"]
        if "organisation" in data_dict:
            del data_dict["organisation"]
        if "last_modified" in data_dict:
            data_dict["last_modified"] = data_dict["last_modified"].split(".")[0]
        if "status" in data_dict:
            if data_dict["status"] == 1:
                data_dict["status"] = "active"
            else:
                data_dict["status"] = "archived"

    def list_contents(self, workspace_id: str):
        """
        This function fetches contents of a workspace from Polly.

        ``Args:``
            |  ``workspace_id :`` workspace id for the target workspace.

        ``Returns:``
            |  it will return a table with attributes.


        .. code::


                # create a obj
                workspaces = Workspaces(token)
                # from there you can other methods
                workspaces.list_contents(workspace_id)
        """
        url = f"{self.base_url}/projects/{workspace_id}/files"
        response = self.session.get(url)
        error_handler(response)
        details_list = []
        data = response.json()["data"]
        columns = ["file_name", "size", "last_modified"]
        for i in data:
            entry_list = []
            file_name = i.get("attributes").get(columns[0])
            size = i.get("attributes").get(columns[1])
            last_modified = i.get("attributes").get(columns[2])
            entry_list.append(file_name)
            entry_list.append(size)
            entry_list.append(last_modified)
            details_list.append(entry_list)
        df = pd.DataFrame(details_list, columns=columns)
        return df

    def create_copy(
        self, source_id: int, source_path: str, destination_id: int, destination_path=""
    ) -> None:
        """
        Function to create a copy of files/folders existing in a workspace into another workspace.

        ``Args:``
            |  ``source_id :`` workspace id of the source workspace where the file/folder exists.
            |  ``source_path :`` file/folder path on the source workspace to be copied.
            |  ``destination_id :`` workspace id of the destination workspace where the file/folder is to be copied.
            |  ``destination_path :`` optional parameter to specify the destination path.

        ``Returns:``
            |  None

        ``Errors:``
            |  ``InvalidParameterException:`` when the parameter like source id is invalid.
            |  ``InvalidPathException:`` when the source path is invalid.


        .. code::


                # create a obj
                workspaces = Workspaces(token)
                # from there you can other methods
                workspaces.create_copy(source_id, source_path, destination_id, destination_path='')

        """
        if not (source_id and isinstance(source_id, int)):
            raise InvalidParameterException("source_id")
        if not (destination_id and isinstance(destination_id, int)):
            raise InvalidParameterException("destination_id")
        if not (source_path and isinstance(source_path, str)):
            raise InvalidParameterException("source_path")
        url = f"{self.base_url}/projects/{destination_id}/files/{destination_path}"
        source_key = f"{source_id}/{source_path}"
        params = {"source": "workspace"}
        sts_url = f"{self.base_url}/projects/{source_id}/credentials/files"
        creds = self.session.get(sts_url)
        error_handler(creds)
        credentials = helpers.get_sts_creds(creds.json())
        bucket = f"mithoo-{self.env_string}-project-data-v1"
        s3_path = f"{bucket}/{source_id}/"
        s3_path = f"s3://{helpers.make_path(s3_path, source_path)}"
        payload = helpers.get_workspace_payload(
            s3_path, credentials, source_key, source_path
        )
        response = self.session.post(url, data=json.dumps(payload), params=params)
        error_handler(response)
        message = response.json()["data"][0].get("attributes", {}).get("body")
        print(message)
        links = response.json().get("included")[0].get("links")
        url = f"{self.base_url}{links.get('self')}"
        while True:
            response = self.session.get(url)
            error_handler(response)
            status = response.json().get("data").get("status")
            if status != "INITIATED":
                break
        print("Copy Operation Successful!")

    def upload_to_workspaces(
        self, workspace_id: int, workspace_path: str, local_path: str
    ) -> None:
        """
        Function to upload files/folders to workspaces.

        ``Args:``
            |  ``workspace_id :`` id of the where file need to uploaded.
            |  ``workspace_path :`` file path on workspace if folder does not exist it will create.
            |  ``local_path :`` uploaded file path.

        ``Returns:``
            |  None

        ``Errors:``
            |  ``InvalidParameterException:`` when the parameter like workspace id is invalid.
            |  ``InvalidPathException:`` when the file to path is invalid.


        .. code::


                # create a obj
                workspaces = Workspaces(token)
                # from there you can other methods
                workspaces.upload_to_workspaces(workspace_id, workspace_path, local_path)

        """
        if not (workspace_id and isinstance(workspace_id, int)):
            raise InvalidParameterException("workspace_id")
        if not (local_path and isinstance(local_path, str)):
            raise InvalidParameterException("local_path")
        if not (workspace_path and isinstance(workspace_path, str)):
            raise InvalidParameterException("workspace_path")
        isExists = os.path.exists(local_path)
        if not isExists:
            raise InvalidPathException
        # check for access rights for the workspace_id
        access_workspace = helpers.workspaces_permission_check(self, workspace_id)
        if not access_workspace:
            raise AccessDeniedError(
                detail=f"Access denied to workspace-id - {workspace_id}"
            )
        sts_url = f"{self.base_url}/projects/{workspace_id}/credentials/files"
        creds = self.session.get(sts_url)
        error_handler(creds)
        credentials = helpers.get_sts_creds(creds.json())
        bucket = f"mithoo-{self.env_string}-project-data-v1"
        s3_path = f"{bucket}/{workspace_id}/"
        s3_path = f"s3://{helpers.make_path(s3_path, workspace_path)}"
        helpers.upload_to_S3(s3_path, local_path, credentials)
        logging.basicConfig(level=logging.INFO)
        logging.info(f"Upload successful on workspace-id={workspace_id}.")

    def download_from_workspaces(self, workspace_id: int, workspace_path: str) -> None:
        """
        Function to download files/folders from workspaces.

        ``Args:``
            |  ``workspace_id :`` id of the where file need to uploaded.
            |  ``workspace_path :`` downloaded file on workspace.

        ``Returns:``
            |  None

        ``Error:``
            |  ``InvalidPathException :`` for invalid path.
            |  ``OperationFailedException :`` when downloading fails.
            |  ``InvalidParameterException:`` when the parameter like workspace id is invalid.


        .. code::


                # create a obj
                workspaces = Workspaces(token)
                # from there you can other methods
                workspaces.download_from_workspaces(workspace_id, workspace_path)

        """
        if not (workspace_id and isinstance(workspace_id, int)):
            raise InvalidParameterException("workspace_id")
        if not (workspace_path and isinstance(workspace_path, str)):
            raise InvalidParameterException("workspace_path")
        # check for access rights for the workspace_id
        access_workspace = helpers.workspaces_permission_check(self, workspace_id)
        if not access_workspace:
            raise AccessDeniedError(
                detail=f"Access denied to workspace-id - {workspace_id}"
            )
        sts_url = f"{self.base_url}/projects/{workspace_id}/credentials/files"
        creds = self.session.get(sts_url)
        error_handler(creds)
        credentials = helpers.get_sts_creds(creds.json())
        bucket = f"mithoo-{self.env_string}-project-data-v1"
        s3_path = f"{bucket}/{workspace_id}/"
        s3_path = f"s3://{helpers.make_path(s3_path, workspace_path)}"
        helpers.download_from_S3(s3_path, workspace_path, credentials)



import base64
import os
import re
import json
import logging
import requests
import urllib.request
from cloudpathlib import S3Client
from botocore.exceptions import ClientError
from cmapPy.pandasGEXpress.parse_gct import parse
from os import environ
from jose import jwt
from polly.errors import (
    BadRequestError,
    error_handler,
    InvalidParameterException,
    MissingKeyException,
    InvalidPathException,
    OperationFailedException,
    paramException,
    AccessDeniedError,
    InvalidRepoException,
)
from polly.constants import COHORT_CONSTANTS_URL, REPORT_FIELDS_URL, ELUCIDATA_LOGO_URL
from bs4 import BeautifulSoup
import contextlib
import joblib
import urllib


def get_platform_value_from_env(
    variable: str, default_val: str, passed_val: str
) -> str:
    """
    Get variable value of passed variable
    from os env variables
    """
    if passed_val:
        default_val = passed_val
    elif environ.get(variable, None) is not None:
        POLLY_TYPE = os.getenv(variable)
        env_val = re.search("https://(.+?).elucidata.io", POLLY_TYPE)
        default_val = env_val.group(1)
    return default_val


def make_path(prefix: any, postfix: any) -> str:
    """
    Function to make and return a valid path
    """
    if not prefix:
        raise InvalidParameterException("prefix")
    if not postfix:
        raise InvalidParameterException("postfix")
    return os.path.normpath(f"{prefix}/{postfix}")


def get_sts_creds(sts_dict: dict) -> dict:
    """
    Function to check and return temporary sts creds
    """
    if sts_dict and isinstance(sts_dict, dict):
        if "data" in sts_dict:
            data = sts_dict.get("data")
            if "attributes" in data[0]:
                attributes = data[0].get("attributes")
                if "credentials" in attributes:
                    return attributes.get("credentials")
                else:
                    raise MissingKeyException("credentials")
            else:
                raise MissingKeyException("attributes")
        else:
            raise MissingKeyException("data")
    else:
        raise InvalidParameterException("sts_dict")


def upload_to_S3(cloud_path: str, local_path: str, credentials: dict) -> None:
    """
    Function to upload file/folder to S3 cloud path
    """
    access_key_id = credentials["AccessKeyId"]
    secret_access_key = credentials["SecretAccessKey"]
    session_token = credentials["SessionToken"]
    client = S3Client(
        aws_access_key_id=access_key_id,
        aws_secret_access_key=secret_access_key,
        aws_session_token=session_token,
    )
    source_path = client.CloudPath(cloud_path)
    if not source_path.exists():
        source_path.mkdir()
    try:
        source_path.upload_from(local_path, force_overwrite_to_cloud=True)
    except ClientError as e:
        raise OperationFailedException(e)


def download_from_S3(cloud_path: str, workspace_path: str, credentials: dict) -> None:
    """
    Function to download file/folder from workspaces
    """
    access_key_id = credentials["AccessKeyId"]
    secret_access_key = credentials["SecretAccessKey"]
    session_token = credentials["SessionToken"]
    client = S3Client(
        aws_access_key_id=access_key_id,
        aws_secret_access_key=secret_access_key,
        aws_session_token=session_token,
    )
    source_path = client.CloudPath(cloud_path)
    if not source_path.exists():
        raise InvalidPathException
    isFile = source_path.is_file()
    if isFile:
        try:
            dest_path = os.getcwd()
            source_path.copy(dest_path, force_overwrite_to_cloud=True)
            logging.basicConfig(level=logging.INFO)
            logging.info(f"Download successful to path={dest_path}")
        except ClientError as e:
            raise OperationFailedException(e)
    else:
        if not cloud_path.endswith("/"):
            cloud_path += "/"
        source_path = client.CloudPath(cloud_path)
        if not source_path.is_dir():
            raise InvalidPathException
        try:
            dest_path = f"{make_path(os.getcwd(),workspace_path)}"
            source_path.copytree(dest_path, force_overwrite_to_cloud=True)
            logging.basicConfig(level=logging.INFO)
            logging.info(f"Download successful to path={dest_path}")
        except ClientError as e:
            raise OperationFailedException(e)


def get_workspace_payload(
    cloud_path: str, credentials: dict, source_key: str, source_path: str
):
    """
    Function to return payload for create_copy function
    """
    access_key_id = credentials["AccessKeyId"]
    secret_access_key = credentials["SecretAccessKey"]
    session_token = credentials["SessionToken"]
    client = S3Client(
        aws_access_key_id=access_key_id,
        aws_secret_access_key=secret_access_key,
        aws_session_token=session_token,
    )
    source_path = client.CloudPath(cloud_path)
    if not source_path.exists():
        raise InvalidPathException
    isFile = source_path.is_file()
    if isFile:
        payload = {
            "data": [
                {
                    "attributes": {
                        "s3_key": source_key,
                    },
                    "id": "",
                    "type": "file",
                }
            ]
        }
    else:
        if not source_key.endswith("/"):
            source_key += "/"
        payload = {
            "data": [
                {
                    "attributes": {
                        "s3_key": source_key,
                    },
                    "id": "",
                    "type": "folder",
                }
            ]
        }
    return payload


def extract_error_title(error_msg: str) -> str:
    """Extract error title from error message

    Args:
        error_msg (str): Whole Error Message in form of string

    Returns:
        str: Return title of the error message
    """
    error_msg = json.loads(error_msg)
    error = error_msg.get("error")
    if error is None:
        error = error_msg.get("errors")[0]
    if "title" in error:
        title = error.get("title")

    return title


def file_conversion(
    self, repo_info: str, dataset_id: str, format: str, header_mapping: dict
) -> None:
    """
    Function that converts file to mentioned format
    """
    if not (repo_info and isinstance(repo_info, str)):
        raise InvalidParameterException("repo_name/repo_id")
    if not (dataset_id and isinstance(dataset_id, str)):
        raise InvalidParameterException("dataset_id")
    if not (format and isinstance(format, str)):
        raise InvalidParameterException("format")
    if not isinstance(header_mapping, dict):
        raise InvalidParameterException("header_mapping")
    download_dict = self.download_data(repo_info, dataset_id)
    url = download_dict.get("data", {}).get("attributes", {}).get("download_url")
    if not url:
        raise MissingKeyException("dataset url")
    file_name = f"{dataset_id}.gct"
    try:
        urllib.request.urlretrieve(url, file_name)
        data = parse(file_name)
        os.remove(file_name)
        row_metadata = data.row_metadata_df
        if header_mapping:
            row_metadata = row_metadata.rename(header_mapping, axis=1)
        row_metadata.to_csv(f"{dataset_id}.{format}", sep="\t")
    except Exception as e:
        raise OperationFailedException(e)


def get_data_type(self, url: str, payload: dict) -> str:
    """
    Function to return the data-type of the required dataset
    """
    if not (url and isinstance(url, str)):
        raise InvalidParameterException("url")
    if not (payload and isinstance(payload, dict)):
        raise InvalidParameterException("payload")
    response = self.session.post(url, data=json.dumps(payload))
    error_handler(response)
    response_data = response.json()
    hits = response_data.get("hits", {}).get("hits")
    if not (hits and isinstance(hits, list)):
        raise paramException(
            title="Param Error",
            detail="No matches found with the given repo details. Please try again.",
        )
    dataset = hits[0]
    data_type = dataset.get("_source", {}).get("data_type")
    if not data_type:
        raise MissingKeyException("data_type")
    return data_type


def get_metadata(self, url: str, payload: dict) -> str:
    """
    Function to return the data-type of the required dataset
    """
    if not (url and isinstance(url, str)):
        raise InvalidParameterException("url")
    if not (payload and isinstance(payload, dict)):
        raise InvalidParameterException("payload")
    response = self.session.post(url, data=json.dumps(payload))
    error_handler(response)
    response_data = response.json()
    hits = response_data.get("hits", {}).get("hits")
    if not (hits and isinstance(hits, list)):
        raise paramException(
            title="Param Error",
            detail="No matches found with the given repo details. Please try again.",
        )
    dataset = hits[0]
    return dataset


def elastic_query(index_name: str, dataset_id: str) -> dict:
    query = {
        "query": {
            "bool": {
                "must": [
                    {"term": {"_index": index_name}},
                    {"term": {"dataset_id.keyword": dataset_id}},
                ]
            }
        }
    }
    return query


def get_cohort_constants() -> json:
    """
    Returns cohort info from public assests url
    """
    response = requests.get(COHORT_CONSTANTS_URL)
    error_handler(response)
    return json.loads(response.text)


def get_cohort_fields() -> json:
    """
    Returns file format info from public assests url
    """
    response = requests.get(REPORT_FIELDS_URL)
    error_handler(response)
    return json.loads(response.text)


def validate_datatype(datatype: str):
    """
    Function to validate datatype of a dataset
    Returns 1 in case of datatype is Single Cell, 0 otherwise
    """
    if datatype == "Single cell":
        return 1
    return 0


@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """Context manager to patch joblib to report into tqdm progress bar given as argument"""

    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()


def check_empty(x):
    """
    Function to validate if the entry is an empty list or not.
    """
    if type(x) == list:
        return len("".join(x))
    elif type(x) == float or type(x) == int:
        return 1
    else:
        return len(x)


def edit_html(local_path: str):
    """
    Function to include Elucidata logo into the report.
    """
    el_image = f"""
        <a>
        <img  style="margin-left: 25px;" src={ELUCIDATA_LOGO_URL}
        width=150" height="70" align="middle">
    </a>
    """
    id_soup = BeautifulSoup(el_image, "html.parser")
    with open(local_path) as fp:
        soup = BeautifulSoup(fp, "html.parser")
    soup.body.insert(1, id_soup)
    with open(local_path, "wb") as f_output:
        f_output.write(soup.prettify("utf-8"))


def verify_workspace_path(cloud_path: str, credentials: dict) -> tuple:
    """
    Function to verify if the workspace path is valid.
    """
    access_key_id = credentials["AccessKeyId"]
    secret_access_key = credentials["SecretAccessKey"]
    session_token = credentials["SessionToken"]
    client = S3Client(
        aws_access_key_id=access_key_id,
        aws_secret_access_key=secret_access_key,
        aws_session_token=session_token,
    )
    source_path = client.CloudPath(cloud_path)
    if source_path.exists():
        return source_path, True
    else:
        return source_path, False


def check_is_file(self, sts_url: str, workspace_id: int, workspace_path: str) -> bool:
    """
    Function to check if the workspace_path is a valid file path existing in the workspace.
    """
    creds = self.session.get(sts_url)
    error_handler(creds)
    credentials = get_sts_creds(creds.json())
    if self.session.env == "polly":
        env_string = "prod"
    elif self.session.env == "testpolly":
        env_string = "test"
    else:
        env_string = "devenv"
    bucket = f"mithoo-{env_string}-project-data-v1"
    s3_path = f"{bucket}/{workspace_id}/"
    s3_path = f"s3://{make_path(s3_path, workspace_path)}"
    tuple_output = verify_workspace_path(s3_path, credentials)
    source_path = tuple_output[0]
    status = tuple_output[1]
    if status is True:
        isFile = source_path.is_file()
        return isFile
    return status


def split_workspace_path(absolute_path: str) -> tuple:
    """
    Function to separate workspace_id and workspace_path from s3 path
    """
    contents = absolute_path.split("/")
    workspace_id = contents[0]
    workspace_path = contents[1]
    for item in range(2, len(contents)):
        workspace_path = make_path(workspace_path, contents[item])
    return workspace_id, workspace_path


def make_private_link(workspace_id: int, workspace_path: str, constant_url: str) -> str:
    """
    Function to construct and return a private link for a file in workspace
    """
    file_element = {"file": f"/projects/{workspace_id}/files/{workspace_path}"}
    private_access_link = f"{constant_url}/workspaces?id={workspace_id}&{urllib.parse.urlencode(file_element)}&preview=true"
    return private_access_link


def change_file_access(
    self, access_key: str, workspace_id: int, workspace_path: str, access_url: str
) -> str:
    """
    Function to change the file access as per the access_key and returns the final access url
    """
    final_url = ""
    if access_key == "private":
        params = {"action": "share", "access_type": "private"}
        url = f"{self.base_url}/projects/{workspace_id}/files/{workspace_path}"
        # API call to change the file access to private
        response = self.session.get(url, params=params)
        error_handler(response)
        final_url = make_private_link(workspace_id, workspace_path, access_url)
    else:
        params = {"action": "share", "access_type": "global"}
        url = f"{self.base_url}/projects/{workspace_id}/files/{workspace_path}"
        # API call to change the file access to public
        response = self.session.get(url, params=params)
        error_handler(response)
        shared_id = response.json().get("data").get("shared_id")
        final_url = f"{access_url}/shared/file/?id={shared_id}"
    return final_url


def get_user_details(self):
    """
    Function to get user details
    """
    me_url = f"{self.base_url}/users/me"
    details = self.session.get(me_url)
    error_handler(details)
    user_details = details.json().get("data", {}).get("attributes")
    return user_details


def get_user_details_using_aws_cognito(self):
    """
    Summary :
    gets the user details such as sub, aud post verification
    using aws cognito

    Returns:
        dictionary of token keys and values
    """
    # getting the userpool details
    pool_url = f"https://api.{self.session.env}.elucidata.io/userpool"
    pool_details = self.session.get(pool_url)
    pool_details_json = pool_details.json()
    user_pool_region = pool_details_json.get("cognito_user_pool_region", "")
    user_pool_id = pool_details_json.get("cognito_user_pool", "")
    # cognito_client_id = pool_details_json.get("cognito_client_id", "")
    session_cookies_dict = self.session.cookies.get_dict()
    id_token = session_cookies_dict.get("idToken", "")
    session_headers_dict = self.session.headers
    cookie = session_headers_dict.get("Cookie", "")
    # cookie is a string
    if "refreshToken=" in cookie:
        refresh_token = cookie.split("refreshToken=")[1]
    else:
        print("refresh token not presen in header cookie")
        refresh_token = ""

    # Getting Headers
    # headers = jwt.get_unverified_header(id_token)
    kid = get_value_from_idToken(id_token, "kid")

    # Getting Pool Url
    verify_url = generate_aws_pool_url(user_pool_region, user_pool_id)

    kargs = {"issuer": verify_url}
    kargs["options"] = {"verify_at_hash": False}
    kargs["audience"] = get_value_from_idToken(id_token, "aud")
    keys = aws_key_dict(user_pool_region, user_pool_id)

    # Get Public Key
    public_key = get_public_key(keys, kid)

    decoded_token = jwt.decode(id_token, public_key, **kargs)
    decoded_token["id_token"] = id_token
    decoded_token["refresh_token"] = refresh_token

    return decoded_token


def generate_aws_pool_url(user_pool_region, user_pool_id):
    verify_url = "https://cognito-idp.{}.amazonaws.com/{}".format(
        user_pool_region, user_pool_id
    )
    return verify_url


def get_public_key(keys, kid):
    """
    Getting public key in pem format from Id Token

    Parameters
    ----------
    keys: Dict
        JWT headers keys
    kid: String
        public key identifier

    Returns
    ----------
    pubk_bytes: String (PEM)
        Public key in pem format
    """

    key = keys[kid]
    return key


def aws_key_dict(region, user_pool_id):
    """
    Fetches the AWS JWT validation file (if necessary) and then converts
    this file into a keyed dictionary that can be used to validate a web-token
    we've been passed

    Parameters
    ----------
    aws_user_pool: String
        AWS Cognito user pool ID
    aws_region: String
        AWS Cognito user pool region

    Returns:
    -------
    dict:
        Contains decoded token dict
    """
    filename = "/tmp/" + "aws_{}.json".format(user_pool_id)

    if not os.path.isfile(filename):
        # If we can't find the file already, try to download it.
        aws_data = requests.get(
            ("https://cognito-idp.{}.amazonaws.com/{}".format(region, user_pool_id))
            + "/.well-known/jwks.json"
        )
        aws_jwt = json.loads(aws_data.text)
        with open(filename, "w+") as json_data:
            json_data.write(aws_data.text)
            json_data.close()

    else:
        with open(filename) as json_data:
            aws_jwt = json.load(json_data)
            json_data.close()

    # We want a dictionary keyed by the kid, not a list.
    result = {}
    for item in aws_jwt["keys"]:
        result[item["kid"]] = item

    return result


def get_value_from_idToken(id_token: str, key: str) -> dict:
    decoded_id_token_dict = decode_id_token(id_token)
    return decoded_id_token_dict.get(key, "")


def decode_id_token(tokenID: str) -> dict:
    """
    Summary: decodes the token id
    this method is to decode the tokenID which consists of 3 parts.
    the first 2 parts can be base64 decoded
    first part contains the kid and alg details.
    second part contains cognito:user details (AWS) such as sub, username
    aud, name, exp, email etc.

    Args:
        tokenId : str

    Returns:
        dict of keys and values from decoded idToken
    """
    decoded_idToken_dict = {}
    try:
        for element in tokenID.split(".")[:2]:
            # the element if lacking "padding" can give 'incorrect padding' error.
            # to handle that we have add the extra padding of '='
            # the len of the element should always be a multiple of 4.
            # https://stackoverflow.com/questions/2941995/python-ignore-incorrect-padding-error-when-base64-decoding

            decoded_str = base64.b64decode(element + "=" * (-len(element) % 4))
            decoded_idToken_dict.update(json.loads(decoded_str.decode("utf-8")))
    except Exception as err:
        print("something wrong with idToken")
        raise err
    return decoded_idToken_dict


def makeRequestCookieForPollyJob(job_data: json) -> json:

    """
    Generates the cookie required request for polly job API

    Args:
        Job data : Dict
    Returns:
        request_cookie: generated cookie string
    """

    secret_env_vars = job_data.get("secret_env", "")

    aud = secret_env_vars.get("POLLY_AUD", "")
    sub = secret_env_vars.get("POLLY_SUB", "")
    idToken = secret_env_vars.get("POLLY_ID_TOKEN", "")
    refreshToken = secret_env_vars.get("POLLY_REFRESH_TOKEN", "")
    idTokenCookie = f"CognitoIdentityServiceProvider.{aud}.{sub}.idToken={idToken}"
    refreshCookie = (
        f"CognitoIdentityServiceProvider.{aud}.{sub}.refreshToken={refreshToken}"
    )
    request_coookie = f"{idTokenCookie};{refreshCookie}"
    # print("returning request_coookie: " + request_coookie)
    return request_coookie


def workspaces_permission_check(self, workspace_id) -> bool:
    """
    Function to check access of a user for a given workspace id.
    """
    permission_url = f"{self.base_url}/workspaces/{workspace_id}/permissions"
    response = self.session.get(permission_url, params={"include": "user"})
    error_handler(response)
    user_details = get_user_details(self)
    user_id = user_details.get("user_id")
    if "data" in response.json():
        json_data = response.json().get("data")
    else:
        raise BadRequestError(detail="Incorrect response format")
    for user in json_data:
        if "attributes" in user:
            attributes = user.get("attributes")
        else:
            raise BadRequestError(detail="Incorrect response format")
        if user_id == attributes["user_id"] and attributes["project_id"] == int(
            workspace_id
        ):
            if attributes["permission"] != "read":
                return True
            else:
                raise AccessDeniedError(
                    detail=f"Read-only permission over the "
                    f"workspace-id {workspace_id}"
                )
    return False


def verify_workspace_details(self, workspace_id, workspace_path, sts_url) -> None:
    """
    Function to check and verify workspace permissions and workspace path.
    """
    access_workspace = workspaces_permission_check(self, workspace_id)
    if not access_workspace:
        raise AccessDeniedError(
            detail=f"Access denied to workspace-id - {workspace_id}"
        )
    is_file = check_is_file(self, sts_url, workspace_id, workspace_path)
    if not is_file:
        raise paramException(
            title="Param Error",
            detail="The given workspace path does not represent a file. Please try again.",
        )


def return_entity_type(data_source: str, cohort_info: json) -> str:
    """
    Function to return entity type based on the cohort info present in public assets
    """
    if data_source not in cohort_info:
        raise InvalidRepoException(data_source)
    for repo, dict in cohort_info.items():
        if data_source == repo:
            if dict["file_structure"] == "single":
                entity_type = "dataset"
            elif dict["file_structure"] == "multiple":
                entity_type = "sample"
    return entity_type


def get_files_in_dir(path_to_dir: str) -> list:
    """
    returns the files in a given directory

    Arguments:
        path_to_dir: str

    Returns:
        list of files in dir : list
    """
    directory = os.fsencode(path_to_dir)
    file_names = os.listdir(directory)
    return file_names


def parseInt(sin):
    """
    parsed the value passed as int  as done by js.
    python equivalent of js parseInt
    example:
        parseInt("100n")  = 100
        parseInt("400m")  = 400

    Arguments:
        sin -- value to be parsed as int

    Returns:
        int
    """
    m = re.search(r"^(\d+)[.,]?\d*?", str(sin))
    parsed_int = int(m.groups()[-1]) if m and not callable(sin) else None
    return parsed_int



from inspect import isclass
import os
import sys
from numpy.core import ufunc
import inspect
from io import StringIO

CACHE = {}


def get_member(name: str, n: str, v: object) -> str:
    # return member name and module name
    try:
        item_name = getattr(v, "__name__", "%s.%s" % (name, n))
        mod_name = getattr(v, "__module__", None)
    except NameError:
        item_name = "%s.%s" % (name, n)
        mod_name = None
    return item_name, mod_name


def is_exe(
    item_name: str, mod_name: str, name: str, n: str, v: object, _all: str
) -> bool:
    # function will return next class or function or modules will push
    # in stack or not
    if "." not in item_name and mod_name:
        item_name = "%s.%s" % (mod_name, item_name)

    if not item_name.startswith(name + "."):
        if isinstance(v, ufunc):
            return False
        else:
            return True
    elif not (inspect.ismodule(v) or _all is None or n in _all):
        return True


def get_docs(name: str, kind: str, index: int, item: object) -> None:
    # storing docstring for a function or class in CACHE variable
    try:
        doc = inspect.getdoc(item)
    except NameError as e:
        raise NameError(title="Name Error", detail=e)

    if doc is not None:
        CACHE[name] = (doc, kind, index)


def path_to_import(this_py: str, mod_path: str, init_py: str) -> str:
    # will return path to import next modules
    if os.path.isfile(this_py) and mod_path.endswith(".py"):
        return mod_path[:-3]
    elif os.path.isfile(init_py):
        return mod_path
    else:
        return None


def get_module(module: str) -> object:
    # importing module
    try:
        __import__(module)
    except ImportError:
        return {}

    return sys.modules[module]


def get_all(item: object) -> object:
    # getting all next items (like function,class) for a module
    try:
        _all = item.__all__
    except AttributeError:
        _all = None
    return _all


def _import(name: str, to_import: str) -> bool:
    # this function will import a given module
    try:
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        try:
            sys.stdout = StringIO()
            sys.stderr = StringIO()
            __import__("%s.%s" % (name, to_import))
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr
            return True
    except (Exception, SystemExit):
        return False


def push_next_members(
    import_modules: bool, item: object, name: str, _all: object, stack
) -> None:
    # pushing next members of given modules
    if import_modules and hasattr(item, "__path__"):
        for pth in item.__path__:
            for mod_path in os.listdir(pth):
                this_py = os.path.join(pth, mod_path)
                init_py = os.path.join(pth, mod_path, "__init__.py")
                to_import = path_to_import(this_py, mod_path, init_py)
                if not to_import or to_import == "__init__":
                    continue
                if not _import(name, to_import):
                    continue

    for n, v in _getmembers(item):
        item_name, mod_name = get_member(name, n, v)
        if is_exe(item_name, mod_name, name, n, v, _all):
            continue
        stack.append(("%s.%s" % (name, n), v))


def _lookfor_generate_cache(module: str, import_modules: bool) -> dict:
    # function will return all the classes and function information
    # like name, docstring, kind for a module
    module = get_module(module)
    seen = {}
    index = 0
    stack = [(module.__name__, module)]

    while stack:
        name, item = stack.pop(0)
        if id(item) in seen:
            continue
        seen[id(item)] = True

        index += 1
        kind = "object"

        if inspect.ismodule(item):
            kind = "module"
            _all = get_all(item)
            push_next_members(import_modules, item, name, _all, stack)
        elif inspect.isclass(item):
            kind = "class"
            for n, v in _getmembers(item):
                stack.append(("%s.%s" % (name, n), v))
        elif hasattr(item, "__call__"):
            kind = "func"
        get_docs(name, kind, index, item)
    return CACHE


def _getmembers(item: object) -> object:
    # returning class member functions
    import inspect

    try:
        members = inspect.getmembers(item)
    except Exception:
        members = [(x, getattr(item, x)) for x in dir(item) if hasattr(item, x)]
    return members


def lookfor(module: str = None, import_modules: bool = True) -> list:
    # function will return names, docstrings  of classes and functions for a module
    CACHE = _lookfor_generate_cache(module, import_modules)

    found = []

    for name, (docstring, kind, index) in CACHE.items():
        if kind in ("module", "object"):
            continue
        doc = docstring.lower()
        found.append((name, kind, doc))

    return found


def checkclass(cls) -> None:
    # checking class or not
    # it is in allowed classes or not
    if not isclass(cls):
        print("Note : use class to get help")
        raise TypeError(title="Use class")

    if cls.__name__ not in ["Polly", "OmixAtlas", "Cohort", "Workspaces"]:
        print("Other class methods not allowed")
        raise Exception(title="Other class are not allowed")


def get_fun_name(function_name: str) -> str:
    # getting function name for input function name
    # get_all_omixatlas() -> get_all_omixatlas
    if function_name and function_name.find("(") != -1:
        return function_name.split("(")[0]
    elif function_name:
        return function_name
    return None


def checkpath(fun: str) -> bool:
    # checking path for a class or function
    # if it includes '_' ,means it is a private function
    # other wise not we should show it
    should_show_fun = True
    for name in fun.split("."):
        if name.startswith("_"):
            should_show_fun = False
            break
    return should_show_fun


def get_line(fun: str, kind: str, txt: str, function_name: str, cls, doc: bool) -> str:
    # function will return lines to print
    # for a function or class
    Link = {
        "cohort": "https://github.com/ElucidataInc/PublicAssets/blob/master/polly-python/example/cohort.ipynb",
        "omixatlas": "https://github.com/ElucidataInc/PublicAssets/blob/master/polly-python/example/omixatlas.ipynb",
        "polly": "https://github.com/ElucidataInc/PublicAssets/blob/master/polly-python/example/polly.ipynb",
        "workspaces": "https://github.com/ElucidataInc/PublicAssets/blob/master/polly-python/example/workspaces.ipynb",
        "add_datasets": "https://github.com/elucidatainc/polly-python/blob/main/ingest/data_ingestion_caseid_1.ipynb",
    }
    line = None
    class_name = cls.__name__.lower()
    description = ""
    lines = txt.split("\n")
    # breaking the text into lines
    for line in lines:
        if not line:
            description = description[:-1]
            break
        description += line + "\n"

    if doc:
        if kind == "class" and not function_name:
            line = (
                "Name - %s()\nDocumentation - %s\n%s class contain following methods\n\n"
                % (fun.split(".")[-1], txt, cls.__name__)
            )
        elif kind == "func" and function_name:
            line = "Name - %s()\nDocumentation - %s \n\n" % (fun.split(".")[-1], txt)
        else:
            line = "Name - %s() \nDescription - %s\n" % (
                fun.split(".")[-1],
                description,
            )
    else:
        if kind == "class" and not function_name:
            line = "Name - %s() \n   Description - %s\n   Link - %s\n" % (
                fun.split(".")[-1],
                description,
                Link[class_name],
            )

        elif kind == "func" and function_name:
            line = "Name - %s() \n   Description - %s\n   Link - %s\n" % (
                fun.split(".")[-1],
                description,
                Link[function_name],
            )

    return line


def get_txt(
    cls,
    fun: str,
    kind: str,
    doc: str,
    function_name: str,
    should_show_fun: bool,
    is_doc: bool,
) -> str:
    # converting function, class docstrings rst to txt
    # returning text to print
    if (
        cls.__name__ in fun.split(".")
        and (not function_name or function_name == fun.split(".")[-1])
        and should_show_fun
    ):

        import rst2txt
        from docutils.core import publish_string

        # converting rst to txt
        txt = publish_string(
            source=doc,
            writer=rst2txt.Writer(),
            settings_overrides={"report_level": 6},
        ).decode("utf-8")

        if txt and "meta private" not in txt:
            while txt.find(">>:ref:`") != -1:
                # removing rst syntax that remains after rst to txt conversion
                # removing link syntax
                x = txt.find(">>:ref:`")
                y = txt.find("`<<") + 3
                str_to_del = txt[x:y]
                str_to_ins = str_to_del.split(">>:ref:`")[1]
                str_to_ins = "visit " + str_to_ins.split(" ")[0]
                txt = txt.replace(str_to_del, str_to_ins)
            line = get_line(fun, kind, txt, function_name, cls, is_doc)
            return line
    return None


def doc(cls, function_name: str = "") -> None:
    """
    function to see documentation for class - Polly, OmixAtlas, Workspaces, Cohort and it's member funtions

    ``Args:``
        ``function_name (optional) str:`` provide function name to see documentation default empty.

    ``Returns:``
        if you provide function_name to search then it will give documentation for function other wise it will\
 give documentation for the class.
    """

    checkclass(cls)
    function_name = get_fun_name(function_name)
    help_text = []
    # pushing those function and classes whose help has been asked
    for (fun, kind, doc) in lookfor("polly", True):
        should_show_fun = checkpath(fun)
        txt = get_txt(cls, fun, kind, doc, function_name, should_show_fun, True)
        if txt:
            help_text.append(txt)
    if len(help_text) == 0:
        help_text.append("Nothing found")
    help_text.append(
        "Go to our polly documentation https://docs.elucidata.io/OmixAtlas/Polly%20Python.html"
    )

    print("\n".join(help_text))


def example(cls, function_name: str = "") -> None:
    """
    function to see examples for class - Polly, OmixAtlas, Workspaces, Cohort and it's member funtions

    ``Args:``
        ``function_name (optional) str:`` provide function name to see examples default empty.

    ``Returns:``
        if you provide function_name to search then it will give example link for that function other wise it \
will give link for all examples for that class.
    """
    checkclass(cls)
    function_name = get_fun_name(function_name)
    help_text = []
    # pushing those function and classes whose examples has been asked
    for (fun, kind, doc) in lookfor("polly", True):
        should_show_fun = checkpath(fun)
        txt = get_txt(cls, fun, kind, doc, function_name, should_show_fun, False)
        if txt:
            help_text.append(txt)

    if help_text == []:
        help_text.append("nothing found")

    print("\n".join(help_text))



from pydantic import BaseModel, constr, create_model, validator, ValidationError, conint
from typing import Optional
import pandas as pd
import copy
from enum import Enum
import re
from tabulate import tabulate
from polly.constants import DDL_CONST_LIST, DML_CONST_LIST, FIELD_NAME_LOC


class FieldType(str, Enum):
    Boolean = "boolean"
    Integer = "integer"
    Float = "float"
    Text = "text"
    Object = "object"

    def __str__(self) -> str:
        return self.value


class AttributeModel(BaseModel):
    original_name: constr(strict=True, min_length=1, max_length=50)
    type: FieldType
    is_keyword: Optional[bool]
    is_array: Optional[bool]
    is_column: Optional[bool]
    is_curated: Optional[bool]
    filter_size: conint(ge=1, le=3000) = 500
    display_name: constr(strict=True, min_length=1, max_length=50)
    description: constr(strict=True, min_length=1, max_length=300)
    is_ontology: Optional[bool]
    is_filter: Optional[bool]

    @validator("is_filter")
    def is_filter_check(cls, is_filter, values):
        if is_filter:
            if not values.get("is_keyword"):
                raise ValueError(
                    f'is_keyword is False and is_filter is True for {values["original_name"]}'
                )
        if values.get("is_ontology", False):
            if not is_filter:
                raise ValueError(
                    f'is_filter is False and is_ontology is True for {values["original_name"]}'
                )
            if not values["is_keyword"]:
                raise ValueError(
                    f'is_keyword is False and is_ontology is True  for {values["original_name"]}'
                )
        return is_filter


def _check_field_name(cls, field_name):
    """
    Check the field name passed in the schema.
    Rules on which it is validated are
    1. Lowercase only
    2. Start with alphabets
    3. Cannot include special characters except `_`
    4. Cannot be longer than 255 characters
    5. Cannot be SQL reserved DDL and DML keywords
    """
    field_name_size = len(field_name)
    if field_name_size < 1 or field_name_size > 255:
        raise ValueError(
            "Size of field name greater than 255 characters or less than 1"
        )

    # regex pattern
    # alphabets character at start
    # no whitespace and special characters in between
    # underscores allowed in between and end
    pattern = r"^[a-z]+[a-z0-9_]*$"
    res = re.match(pattern, field_name)
    if not res:
        raise ValueError(
            f"The field name {field_name} did not match the defined naming conditions. "
            + "Lowercase only, Start with alphabets,  Cannot include special characters except `_`, "
            + "Cannot be longer than 255 characters, Cannot be SQL reserved DDL and DML keywords."
        )

    ddl_consts = [x.lower() for x in DDL_CONST_LIST]
    dml_consts = [x.lower() for x in DML_CONST_LIST]

    if field_name in ddl_consts:
        raise ValueError(f"The field name {field_name} matches reserved DDL Constants.")

    if field_name in dml_consts:
        raise ValueError(f"The field name {field_name} matches reserved DML Constants.")

    return field_name


def _build_schema_validation_model(field_name: str, validators):
    """
    Return a schema validation model using pydantic create model
    """
    schema_model = create_model(
        "SchemaValidation",
        field_name=(str, ...),
        __validators__=validators,
        __base__=AttributeModel,
    )
    return schema_model


def _validate_schema_values(field_dict: dict):
    """
    Create pydantic model to validate schema values
    """
    field_name = field_dict["field_name"]
    validators = {
        "check_field_name": validator(
            "field_name", allow_reuse=True, check_fields=False
        )(_check_field_name)
    }
    schema_model = _build_schema_validation_model(field_name, validators)
    try:
        schema_model(**field_dict)
    except ValidationError as s_errors:
        error_data_list = []
        for error in s_errors.errors():
            error_data = {"field_name": field_name}
            # getting 1st element from tuple loc
            # which contains location of the error
            if len(error.get("loc")) > 0:
                error_loc = error.get("loc")[0]
            else:
                error_loc = ""
            if error_loc == FIELD_NAME_LOC:
                error_data["attribute"] = "Not Applicable"
            else:
                attribute_name = error_loc
                error_data["attribute"] = attribute_name
            error_data["message"] = error.get("msg")
            error_data_list.append(error_data)
        schema_errors = pd.DataFrame(error_data_list)
        return schema_errors


def check_schema_for_errors(fields_dict):
    """
    Check list of field dictionaries for possible errors
    """
    errors_all = []
    for field_dict_key, field_dict_attributes in fields_dict.items():
        field_dict = {}
        field_dict = copy.deepcopy(field_dict_attributes)
        field_dict["field_name"] = field_dict_key
        errors = _validate_schema_values(field_dict)
        if errors is not None:
            errors_all.append(errors)

    if errors_all:
        error_df = pd.concat(errors_all, ignore_index=True)
        print(tabulate(error_df, headers="keys", tablefmt="fancy_grid"))
        return error_df
    else:
        print("Schema Passed Has No errors")
        error_df = pd.DataFrame()
        return error_df



DATA_TYPES = {
    "Mutation": [
        {
            "format": ["maf"],
            "supported_repo": [
                {
                    "name": "cbioportal",
                    "header_mapping": {
                        "gene": "Hugo_Symbol",
                        "chr": "Chromosome",
                        "startPosition": "Start_Position",
                        "endPosition": "End_Position",
                        "referenceAllele": "Reference_Allele",
                        "variantAllele": "Tumor_Seq_Allele2",
                        "mutationType": "Variant_Classification",
                        "variantType": "Variant_Type",
                        "uniqueSampleKey": "Tumor_Sample_Barcode",
                    },
                },
                {"name": "tcga", "header_mapping": {}},
            ],
        }
    ]
}

# endpoints
CONSTANTS_ENDPOINT = "/constants"
REPOSITORIES_ENDPOINT = "/repositories"
REPOSITORY_PACKAGE_ENDPOINT = REPOSITORIES_ENDPOINT + "/{}/packages"
IMAGE_URL_ENDPOINT = (
    "https://elucidatainc.github.io/PublicAssets/discover-fe-assets/omixatlas_hex.svg"
)

# statuscodes
OK = 200
CREATED = 201
COMPUTE_ENV_VARIABLE = "POLLY_TYPE"
UPLOAD_URL_CREATED = 204
UPLOAD_ERROR_CODE = 400

# Multipart S3 upload parameters
KB = 1024
MB = KB * KB
GB = 1024 * MB
SMALL_FILE_SIZE = 20 * GB
MEDIUM_FILE_SIZE = 40 * GB

MULTIPART_THRESHOLD = 25 * MB
MAX_CONCURRENCY = 32

# Chunksize for file less than 20 gb
MULTIPART_CHUNKSIZE_SMALL_FILE_SIZE = 25 * MB
IO_CHUNKSIZE_SMALL_FILE_SIZE = 25 * MB

# Chunksize for file more than 20gb and less than 40gb
MULTIPART_CHUNKSIZE_MEDIUM_FILE_SIZE = 50 * MB
IO_CHUNKSIZE_MEDIUM_FILE_SIZE = 50 * MB

# Chunksize for file more than 40gb
MULTIPART_CHUNKSIZE_LARGE_FILE_SIZE = 100 * MB
IO_CHUNKSIZE_LARGE_FILE_SIZE = 100 * MB

# S3 Exceptions
EXPIRED_TOKEN = "ExpiredToken"

# cohort constants
COHORT_VERSION = "0.2"
COHORT_CONSTANTS_URL = (
    "https://elucidatainc.github.io/PublicAssets/cohort_constants.txt"
)
REPORT_FIELDS_URL = "https://elucidatainc.github.io/PublicAssets/report_fields.txt"
OBSOLETE_METADATA_FIELDS = [
    "package",
    "region",
    "bucket",
    "key",
    "file_type",
    "file_location",
    "src_uri",
    "timestamp_",
]
dot = "."

GETTING_UPLOAD_URLS_PAYLOAD = {"data": {"type": "files", "attributes": {"folder": ""}}}

INGESTION_LEVEL_METADATA = {
    "id": "metadata/ingestion",
    "type": "ingestion_metadata",
    "attributes": {
        "ignore": "false",
        "urgent": "true",
        "v1_infra": False,
        "priority": "low",
    },
}

METADATA = {"data": []}

GET_SCHEMA_RETURN_TYPE_VALS = ["dataframe", "dict"]

COMBINED_METADATA_FILE_NAME = "combined_metadata.json"

FILES_PATH_FORMAT = {"metadata": "<metadata_path>", "data": "<data_path>"}

INGESTION_FILES_PATH_DIR_NAMES = ["metadata", "data"]

FORMATTED_METADATA = {"id": "", "type": "", "attributes": {}}

FILE_FORMAT_CONSTANTS_URL = (
    "https://elucidatainc.github.io/PublicAssets/file_format_constants.txt"
)

COMPRESSION_TYPES = [
    ".br," ".bz2",
    ".bz",
    ".gz",
    ".lz",
    ".lz4",
    ".sz",
    ".rz",
    ".xz",
    ".zip",
    ".tar",
    ".bgz",
]

FILE_FORMAT_CONSTANTS = {
    "data": [
        ".gct",
        ".vcf",
        ".h5ad",
        ".mmcif",
        ".h5seurat",
        ".biom",
        ".zip",
        ".fs",
        ".tar.gz",
        ".fcs",
        ".vcf.bgz",
        ".gct.bz",
    ],
    "metadata": [".json", ".jpco"],
}
NOT_NEEDED_SCHEMA_FIELDS = [
    "is_current",
    "data_table_version",
    "data_table_name",
    "timestamp_",
]

VALIDATION_LEVEL_CONSTANTS = {"advanced": "value", "basic": "schema"}

BASE_TEST_FORMAT_CONSTANTS_URL = "https://github.com/ElucidataInc/PublicAssets/blob/master/internal-user/add_dataset_test_file"

BASE_TEST_META_CONST = "https://raw.githubusercontent.com/ElucidataInc/PublicAssets/master/internal-user/add_dataset_test_file"

ELUCIDATA_LOGO_URL = (
    "https://elucidatainc.github.io/PublicAssets/dashboardFrontend/elucidata-logo.svg"
)

SCHEMA_WRITE_SUCCESS = (
    "Job for your schema update has been recorded. Depending on the changes,"
    + "minor updates or complete reindexing will be performed for you data. The repository will be locked for the "
    + "duration of this update. Please refer to the ingestion monitoring dashboard for visibility on this process."
)


REPORT_GENERATION_SUPPORTED_REPOS = ["geo"]

FIELD_NAME_LOC = "field_name"

DDL_CONST_LIST = [
    "ALL",
    "ALTER",
    "AND",
    "ARRAY",
    "AS",
    "AUTHORIZATION",
    "BETWEEN",
    "BIGINT",
    "BINARY",
    "BOOLEAN",
    "BOTH",
    "BY",
    "CASE",
    "CASHE",
    "CAST",
    "CHAR",
    "COLUMN",
    "CONF",
    "CONSTRAINT",
    "COMMIT",
    "CREATE",
    "CROSS",
    "CUBE",
    "CURRENT",
    "CURRENT_DATE",
    "CURRENT_TIMESTAMP",
    "CURSOR",
    "DATABASE",
    "DATE",
    "DAYOFWEEK",
    "DECIMAL",
    "DELETE",
    "DESCRIBE",
    "DISTINCT",
    "DOUBLE",
    "DROP",
    "ELSE",
    "END",
    "EXCHANGE",
    "EXISTS",
    "EXTENDED",
    "EXTERNAL",
    "EXTRACT",
    "FALSE",
    "FETCH",
    "FLOAT",
    "FLOOR",
    "FOLLOWING",
    "FOR",
    "FOREIGN",
    "FROM",
    "FULL",
    "FUNCTION",
    "GRANT",
    "GROUP",
    "GROUPING",
    "HAVING",
    "IF",
    "IMPORT",
    "IN",
    "INNER",
    "INSERT",
    "INT",
    "INTEGER",
    "INTERSECT",
    "INTERVAL",
    "INTO",
    "IS",
    "JOIN",
    "LATERAL",
    "LEFT",
    "LESS",
    "LIKE",
    "LOCAL",
    "MACRO",
    "MAP",
    "MORE",
    "NONE",
    "NOT",
    "NULL",
    "NUMERIC",
    "OF",
    "ON",
    "ONLY",
    "OR",
    "ORDER",
    "OUT",
    "OUTER",
    "OVER",
    "PARTIALSCAN",
    "PARTITION",
    "PERCENT",
    "PRECEDING",
    "PRECISION",
    "PRESERVE",
    "PRIMARY",
    "PROCEDURE",
    "RANGE",
    "READS",
    "REDUCE",
    "REGEXP",
    "REFERENCES",
    "REVOKE",
    "RIGHT",
    "RLIKE",
    "ROLLBACK",
    "ROLLUP",
    "ROW",
    "ROWS",
    "SELECT",
    "SET",
    "SMALLINT",
    "START",
    "TABLE",
    "TABLESAMPLE",
    "THEN",
    "TIME",
    "TIMESTAMP",
    "TO",
    "TRANSFORM",
    "TRIGGER",
    "TRUE",
    "TRUNCATE",
    "UNBOUNDED",
    "UNION",
    "UNIQUEJOIN",
    "UPDATE",
    "USER",
    "USING",
    "UTC_TIMESTAMP",
    "VALUES",
    "VARCHAR",
    "VIEWS",
    "WHEN",
    "WHERE",
    "WINDOW",
    "WITH",
]

DML_CONST_LIST = [
    "ALTER",
    "AND",
    "AS",
    "BETWEEN",
    "BY",
    "CASE",
    "CAST",
    "CONSTRAINT",
    "CREATE",
    "CROSS",
    "CUBE",
    "CURRENT_DATE",
    "CURRENT_PATH",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "CURRENT_USER",
    "DEALLOCATE",
    "DELETE",
    "DESCRIBE",
    "DISTINCT",
    "DROP",
    "ELSE",
    "END",
    "ESCAPE",
    "EXCEPT",
    "EXECUTE",
    "EXISTS",
    "EXTRACT",
    "FALSE",
    "FIRST",
    "FOR",
    "FROM",
    "FULL",
    "GROUP",
    "GROUPING",
    "HAVING",
    "IN",
    "INNER",
    "INSERT",
    "INTERSECT",
    "INTO",
    "IS",
    "JOIN",
    "LAST",
    "LEFT",
    "LIKE",
    "LOCALTIME",
    "LOCALTIMESTAMP",
    "NATURAL",
    "NORMALIZE",
    "NOT",
    "NULL",
    "OF",
    "ON",
    "OR",
    "ORDER",
    "OUTER",
    "PREPARE",
    "RECURSIVE",
    "RIGHT",
    "ROLLUP",
    "SELECT",
    "TABLE",
    "THEN",
    "TRUE",
    "UNESCAPE",
    "UNION",
    "UNNEST",
    "USING",
    "VALUES",
    "WHEN",
    "WHERE",
    "WITH",
]

# curation_library constants
SUPPORTED_ENTITY_TYPES = [
    "disease",
    "drug",
    "species",
    "tissue",
    "cell_type",
    "cell_line",
    "gene",
]

SCHEMA_VALIDATION_BASE_URL = (
    "https://raw.githubusercontent.com/ElucidataInc/PublicAssets/master/"
    + "internal-user/schema_validation"
)

SCHEMA_VALIDATION = {
    "wrong_repo_id": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_wrong_repo_id.json",
    "field_name_cap": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_field_name_capital.json",
    "field_name_underscore": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_field_name_with_underscore.json",
    "field_name_resv_keyword": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_field_name_with_resv_keyword.json",
    "original_name_empty": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_original_name_empty.json",
    "original_name_grtr_50": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_original_name_gtr_50.json",
    "type_cosco": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_type_cosco.json",
    "is_arr_str": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_is_arr_str.json",
    "is_keyword_str": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_is_keywrd_str.json",
    "filter_size_less": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_filter_size_less.json",
    "filter_size_greater": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_filter_size_greater.json",
    "display_name_empty": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_display_name_empty.json",
    "display_name_grtr_50": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_display_name_grtr_50.json",
    "is_keywrd_is_filter": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_is_keywrd_is_filter.json",
    "is_keywrd_is_ontology": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_field_name_is_filter_is_ontology.json",
}

MIN_REQUIRED_KEYS_FOR_JOBS = ["cpu", "memory", "machineType", "image", "tag", "name"]

MACHINES_FOR_JOBS = (
    {
        "gp": "4 vCPU, 16GB RAM",
        "ci2xlarge": "16 vCPU, 32GB RAM",
        "ci3xlarge": "36 vCPU, 72GB RAM",
        "mi2xlarge": "4 vCPU, 32GB RAM",
        "mi3xlarge": "8 vCPU, 64GB RAM",
        "mi4xlarge": "16 vCPU, 122GB RAM",
        "mi5xlarge": "32 vCPU, 250GB RAM",
        "mi6xlarge": "64 vCPU, 500GB RAM",
        "mi7xlarge": "64 vCPU, 970GB RAM",
        "mix5xlarge": "16vCPU, 512GB RAM",
        "mix6xlarge": "24vCPU, 768GB RAM",
        "mix7xlarge": "64vCPU, 1024GB RAM",
        "gpusmall": "1 GPU, 4 vCPU, 16GB RAM",
        "gpumedium": "4 GPU, 32 vCPU, 240GB RAM",
        "gpularge": "8 GPU, 64 vCPU, 480GB RAM",
        "gpuxlarge": "8 GPU, 96 vCPU, 760GB RAM",
    },
)



import pathlib
import os
from setuptools import setup, find_packages

UPSTREAM_URLLIB3_FLAG = "--with-upstream-urllib3"


def get_requirements(raw=False):
    """Build the requirements list for this project"""
    requirements_list = []

    with open("requirements.txt") as reqs:
        for install in reqs:
            if install.startswith("# only telegram.ext:"):
                if raw:
                    break
                continue
            requirements_list.append(install.strip())

    return requirements_list


def get_packages_requirements(raw=False):
    """Build the package & requirements list for this project"""
    reqs = get_requirements(raw=raw)
    exclude = ["tests*"]
    packs = find_packages(exclude=exclude)
    return packs, reqs


packages, requirements = get_packages_requirements()
# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()


def read(rel_path: str) -> str:
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path)) as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


# This call to setup() does all the work
# circle bracket and format in the download_url parameter to resolve linting issue
# of line too long
setup(
    name="polly-python",
    version=get_version("polly/__init__.py"),
    description="Polly SDK",
    long_description=README,
    long_description_content_type="text/markdown",
    packages=packages,
    include_package_data=True,
    setup_requires=["wheel"],
    install_requires=requirements,
    url="https://github.com/ElucidataInc/polly-python",
    download_url=(
        "https://elucidatainc.github.io/PublicAssets/builds/polly-python/"
        "polly_python-{a}-none-any.whl".format(a=get_version("polly/__init__.py"))
    ),
)
